# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""youtube-vis dataset."""

from __future__ import annotations

import collections
import json
import os
from typing import Any, Dict, List, Optional, Tuple, Union

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Youtube-vis is a video instance segmentation dataset. It contains 2,883
high-resolution YouTube videos, a per-pixel category label set including 40
common objects such as person, animals and vehicles, 4,883 unique video
instances, and 131k high-quality manual annotations.

The YouTube-VIS dataset is split into 2,238 training videos, 302 validation
videos and 343 test videos.

No files were removed or altered during preprocessing.
"""

_CITATION = """
@article{DBLP:journals/corr/abs-1905-04804,
  author    = {Linjie Yang and
               Yuchen Fan and
               Ning Xu},
  title     = {Video Instance Segmentation},
  journal   = {CoRR},
  volume    = {abs/1905.04804},
  year      = {2019},
  url       = {http://arxiv.org/abs/1905.04804},
  archivePrefix = {arXiv},
  eprint    = {1905.04804},
  timestamp = {Tue, 28 May 2019 12:48:08 +0200},
  biburl    = {https://dblp.org/rec/journals/corr/abs-1905-04804.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

NUM_TRAIN_EXAMPLES = 2238

NestedDict = Dict[str, Any]


def _convert_bbox(
    box: List[float], height: int, width: int
) -> tfds.features.BBox:
  """Converts bbox from coco x,y,w,h to xmin, ymin, xmax, ymax tfds format."""
  return tfds.features.BBox(
      xmin=box[0] / width,
      ymin=box[1] / height,
      xmax=(box[0] + box[2]) / width,
      ymax=(box[1] + box[3]) / height,
  )  # pytype: disable=bad-return-type  # gen-stub-imports


def _decode_segmentation(
    segmentation: Union[List[NestedDict], NestedDict],
    video: NestedDict,
    desired_height: int,
    desired_width: int,
):
  """Converts the run length encoded segmentation into an image."""
  pycocotools = tfds.core.lazy_imports.pycocotools
  rle = pycocotools.frPyObjects(segmentation, video['height'], video['width'])
  if isinstance(segmentation, list):  # Polygon
    rle = pycocotools.merge(rle)
  segmentation = pycocotools.decode(rle)
  assert segmentation.shape[0] == video['height']
  assert segmentation.shape[1] == video['width']
  if video['height'] != desired_height or video['width'] != desired_width:
    cv2 = tfds.core.lazy_imports.cv2
    segmentation = cv2.resize(
        segmentation,
        (desired_width, desired_height),
        interpolation=cv2.INTER_NEAREST,
    )
  segmentation = np.expand_dims(segmentation, axis=-1)
  assert len(segmentation.shape) == 3
  return segmentation


def _find_frame_index(
    frame_filename: str, all_video_frame_paths: List[epath.PathLike]
) -> int:
  for index, path in enumerate(all_video_frame_paths):
    if frame_filename in os.fspath(path):
      return index
  assert False, (
      'Annotations are corrupt or videos have not been properly '
      f'downloaded. File {frame_filename} not found in '
      f'{all_video_frame_paths}.'
  )


def _create_per_track_annotation(
    video: NestedDict,
    all_video_frame_paths: List[epath.PathLike],
    track_annotation: NestedDict,
    desired_height: Optional[int] = None,
    desired_width: Optional[int] = None,
) -> NestedDict:
  """Creates an anntation for a single object track.

  Args:
    video: The annotation for the video containing metadata info.
    all_video_frame_paths: List of all the frames belonging to the video.
    track_annotation: The annotation for a single object track in the video.
    desired_height: The height the video will be resized to. If None, no
      resizing occurs.
    desired_width: The width the video will be resized to. If None, no resizing.
      occurs.

  Returns:
    A data entry for a single object track for the video.
  """
  assert video['id'] == track_annotation['video_id']
  height = desired_height or video['height']
  width = desired_width or video['width']
  per_track_anno = {}

  per_track_anno['bboxes'] = []  # Temporally ordered bounding boxes for track.
  per_track_anno['frames'] = []  # List of frames the track appears on.
  frames_with_labels = []
  for frame_idx, box in enumerate(track_annotation['bboxes']):
    if box is None:
      continue
    frames_with_labels.append(frame_idx)
    per_track_anno['bboxes'].append(
        _convert_bbox(box, video['height'], video['width'])
    )
    # all_video_frame_paths is a superset of the annotated frames, and we
    # need to convert the index into the annotated frames to an index into
    # all_video_frame_paths.
    per_track_anno['frames'].append(
        _find_frame_index(video['file_names'][frame_idx], all_video_frame_paths)
    )
  frames_with_labels = set(frames_with_labels)

  per_track_anno['segmentations'] = []  # Temporally ordered segmentations.
  for frame_idx, segmentation in enumerate(track_annotation['segmentations']):
    if segmentation is None:
      assert frame_idx not in frames_with_labels
      continue
    per_track_anno['segmentations'].append(
        _decode_segmentation(segmentation, video, height, width)
    )
    assert frame_idx in frames_with_labels

  per_track_anno['areas'] = []  # List of per-pixel segmentation areas.
  for frame_idx, area in enumerate(track_annotation['areas']):
    if area is None:
      assert frame_idx not in frames_with_labels
      continue
    size_ratio = float(video['height'] * video['width']) / (height * width)
    area = int(size_ratio * area)
    per_track_anno['areas'].append(area)
    assert frame_idx in frames_with_labels

  # TFDS ids are zero indexed whereas youtube-vis ids are not.
  per_track_anno['category'] = track_annotation['category_id'] - 1
  per_track_anno['is_crowd'] = track_annotation['iscrowd']

  return per_track_anno


def _create_metadata(
    video: NestedDict, height: int, width: int, num_frames: int
) -> NestedDict:
  """Creates the metadata entry for a video."""
  metadata = {}
  metadata['height'] = height or video['height']
  metadata['width'] = width or video['width']
  metadata['num_frames'] = num_frames
  metadata['video_name'] = video['file_names'][0].split('/')[0]
  return metadata


def _build_annotations_index(
    annotations: NestedDict,
) -> Tuple[Dict[int, List[NestedDict]], Dict[int, NestedDict]]:
  """Builds some indices to make data generation more convenient."""
  video_id_to_annos = collections.defaultdict(list)
  videos = {}
  if 'annotations' in annotations:
    for anno in annotations['annotations']:
      video_id_to_annos[anno['video_id']].append(anno)
  else:  # Testing or validation data contains no annotations.
    video_id_to_annos = {v['id']: [] for v in annotations['videos']}
  videos = {v['id']: v for v in annotations['videos']}
  return video_id_to_annos, videos


class YoutubeVisConfig(tfds.core.BuilderConfig):
  """ "Configuration for Youtube-vis video instance segmentation dataset.

  Attributes:
    height: An optional integer height to resize all videos to. If None, no
      resizing will occur.
    width: An optional integer width to resize all videos to. If None, no
      resizing will occur.
    only_frames_with_labels: A bool indicating whether we should include only
      frames which have labels (True) or whether all frames, including those
      without labels, should be included (False).
  """

  def __init__(
      self,
      *,
      height: Optional[int] = None,
      width: Optional[int] = None,
      only_frames_with_labels: bool = False,
      split_train_data_range: Optional[Tuple[int, int]] = None,
      split_val_data_range: Optional[Tuple[int, int]] = None,
      split_test_data_range: Optional[Tuple[int, int]] = None,
      **kwargs,
  ):
    """The parameters specifying how the dataset will be processed.

    This allows the option to preprocess the images to a smaller fixed
    resolution. If height and width are left as None, the native size
    will be used. It also allows dropping all frames which don't include
    labels.

    Args:
      height: optional height to resize all images to
      width: optional width to resize all images to
      only_frames_with_labels: whether or not to include the frames which don't
        have labels in the data export.
      split_train_data_range: If not None, must be a tuple of two integers
        indicating the slice (left-inclusive, right-exclusive) of the train data
        to subsample into a manufactured training split. Because the default
        validation and test splits contain no labels, this allows for validation
        and testing without uploading to the youtube_vis third party servers.
      split_val_data_range: If not None, must be a tuple of integers indicating
        the slice (left-inclusive, right-exclusive) of the train data to
        subsample into a manufactured validation split. Because the default
        validation and test splits contain no labels, this allows for validation
        and testing without uploading to the youtube_vis third party servers.
      split_test_data_range: If not None, must be a tuple of integers indicating
        the slice (left-inclusive, right-exclusive) of the train data to
        subsample into a manufactured test split. Because the default validation
        and test splits contain no labels, this allows for validation and
        testing without uploading to the youtube_vis third party servers.
      **kwargs: Passed on to the constructor of `BuilderConfig`.

    Raises:
      ValueError if split data ranges are outside of acceptable values.
    """
    super(YoutubeVisConfig, self).__init__(**kwargs)
    if height is not None or width is not None:
      if height is None or width is None:
        raise ValueError('Provide either both height and width or None.')
    self.height = height
    self.width = width
    self.only_frames_with_labels = only_frames_with_labels
    # Check that the requested data falls within the allowed ranges.
    # NOTE(austinstone): We do not check that the split data is non-overlapping.
    # The user should ensure that their splits are not-overlapping to prevent
    # training on the test data.
    if split_train_data_range is not None:
      if (
          split_train_data_range[0] < 0
          or split_train_data_range[1] > NUM_TRAIN_EXAMPLES
      ):
        raise ValueError(
            f'split_train_data_range must be within [0, {NUM_TRAIN_EXAMPLES}] ',
            f'got instead: {split_train_data_range}.',
        )
    if split_val_data_range is not None:
      if (
          split_val_data_range[0] < 0
          or split_val_data_range[1] > NUM_TRAIN_EXAMPLES
      ):
        raise ValueError(
            f'split_val_data_range must be within [0, {NUM_TRAIN_EXAMPLES}] ',
            f'got instead: {split_val_data_range}.',
        )
    if split_test_data_range is not None:
      if (
          split_test_data_range[0] < 0
          or split_test_data_range[1] > NUM_TRAIN_EXAMPLES
      ):
        raise ValueError(
            f'split_test_data_range must be within [0, {NUM_TRAIN_EXAMPLES}] ',
            f'got instead: {split_test_data_range}.',
        )
    self.split_train_data_range = split_train_data_range
    self.split_val_data_range = split_val_data_range
    self.split_test_data_range = split_test_data_range


class YoutubeVis(tfds.core.BeamBasedBuilder):
  """DatasetBuider for Youtube-vis dataset."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Please download all files for the 2019 version of the dataset
  (test_all_frames.zip, test.json, train_all_frames.zip, train.json,
  valid_all_frames.zip, valid.json) from the youtube-vis website
  and move them to ~/tensorflow_datasets/downloads/manual/.

  Note that the dataset landing page is located at
  https://youtube-vos.org/dataset/vis/, and it will then redirect you to a page
  on https://competitions.codalab.org where you can download the 2019 version
  of the dataset. You will need to make an account on codalab to download the
  data. Note that at the time of writing this, you will need to bypass a
  "Connection not secure" warning when accessing codalab.
  """
  BUILDER_CONFIGS = [
      YoutubeVisConfig(
          name='full',
          description=(
              'The full resolution version of the dataset, with all '
              'frames, including those without labels, included.'
          ),
      ),
      YoutubeVisConfig(
          name='480_640_full',
          description=(
              'All images are bilinearly resized to 480 X 640 with all '
              'frames included.'
          ),
          height=480,
          width=640,
      ),
      YoutubeVisConfig(
          name='480_640_only_frames_with_labels',
          description=(
              'All images are bilinearly resized to 480 X 640 with only'
              ' frames with labels included.'
          ),
          height=480,
          width=640,
          only_frames_with_labels=True,
      ),
      YoutubeVisConfig(
          name='only_frames_with_labels',
          description=(
              'Only images with labels included at their native resolution.'
          ),
          only_frames_with_labels=True,
      ),
      # BEGIN GOOGLE_INTERNAL
      # These splits will be released to third party consumers after internal
      # testing.
      YoutubeVisConfig(
          name='full_train_split',
          description=(
              'The full resolution version of the dataset, with all frames,'
              ' including those without labels, included. The val and test'
              ' splits are manufactured from the training data.'
          ),
          # Use the first 1838 train videos for training.
          split_train_data_range=(0, 1838),
          # Use training videos 1838-2038 for validation.
          split_val_data_range=(1838, 2038),
          # Use training videos 2038-2238 for testing.
          split_test_data_range=(2038, 2238),
      ),
      YoutubeVisConfig(
          name='480_640_full_train_split',
          description=(
              'All images are bilinearly resized to 480 X 640 with all '
              'frames included. The val and test splits are '
              'manufactured from the training data.'
          ),
          height=480,
          width=640,
          # Use the first 1838 train videos for training.
          split_train_data_range=(0, 1838),
          # Use training videos 1838-2038 for validation.
          split_val_data_range=(1838, 2038),
          # Use training videos 2038-2238 for testing.
          split_test_data_range=(2038, 2238),
      ),
      YoutubeVisConfig(
          name='480_640_only_frames_with_labels_train_split',
          description=(
              'All images are bilinearly resized to 480 X 640 with only'
              ' frames with labels included. The val and test splits '
              'are manufactured from the training data.'
          ),
          height=480,
          width=640,
          only_frames_with_labels=True,
          # Use the first 1838 train videos for training.
          split_train_data_range=(0, 1838),
          # Use training videos 1838-2038 for validation.
          split_val_data_range=(1838, 2038),
          # Use training videos 2038-2238 for testing.
          split_test_data_range=(2038, 2238),
      ),
      YoutubeVisConfig(
          name='only_frames_with_labels_train_split',
          description=(
              'Only images with labels included at their native '
              'resolution. The val and test splits are manufactured '
              'from the training data.'
          ),
          only_frames_with_labels=True,
          # Use the first 1838 train videos for training.
          split_train_data_range=(0, 1838),
          # Use training videos 1838-2038 for validation.
          split_val_data_range=(1838, 2038),
          # Use training videos 2038-2238 for testing.
          split_test_data_range=(2038, 2238),
      ),
      # END GOOGLE_INTERNAL
  ]
  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    names_file = tfds.core.tfds_path('video/youtube_vis/labels.txt')
    video_shape = (
        None,
        self.builder_config.height,
        self.builder_config.width,
        3,
    )
    seg_shape = (None, self.builder_config.height, self.builder_config.width, 1)
    all_features = {
        'video': tfds.features.Video(video_shape),  # pytype: disable=wrong-arg-types  # gen-stub-imports
        'metadata': {
            'height': np.int32,
            'width': np.int32,
            'num_frames': np.int32,
            'video_name': np.str_,
        },
        'tracks': tfds.features.Sequence({
            'bboxes': tfds.features.Sequence(tfds.features.BBoxFeature()),
            'segmentations': tfds.features.Video(seg_shape, use_colormap=True),  # pytype: disable=wrong-arg-types  # gen-stub-imports
            'category': tfds.features.ClassLabel(names_file=names_file),
            'is_crowd': np.bool_,
            'areas': tfds.features.Sequence(np.float32),
            # Labels do not occur for all frames. This indicates the
            # indices of the frames that have labels.
            'frames': tfds.features.Sequence(np.int32),
        }),
    }
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(all_features),
        supervised_keys=None,
        homepage='https://youtube-vos.org/dataset/vis/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    manually_downloaded_files = {
        'train_all_frames': dl_manager.manual_dir / 'train_all_frames.zip',
        'train_annotations': dl_manager.manual_dir / 'train.json',
    }
    if self.builder_config.split_train_data_range is not None:
      # Create a custom training split by subsampling the training data.
      train_data_range = self.builder_config.split_train_data_range
    else:  # Use the provided training split.
      train_data_range = None

    if self.builder_config.split_val_data_range is not None:
      # Create a custom validation split by subsampling the training data.
      val_data_range = self.builder_config.split_val_data_range
      manually_downloaded_files['valid_all_frames'] = manually_downloaded_files[
          'train_all_frames'
      ]
      manually_downloaded_files['valid_annotations'] = (
          manually_downloaded_files['train_annotations']
      )
    else:  # Use the provided validation split.
      val_data_range = None
      manually_downloaded_files['valid_all_frames'] = (
          dl_manager.manual_dir / 'valid_all_frames.zip'
      )
      manually_downloaded_files['valid_annotations'] = (
          dl_manager.manual_dir / 'valid.json'
      )

    if self.builder_config.split_test_data_range is not None:
      # Create a custom test split by subsampling the training data.
      test_data_range = self.builder_config.split_test_data_range
      manually_downloaded_files['test_all_frames'] = manually_downloaded_files[
          'train_all_frames'
      ]
      manually_downloaded_files['test_annotations'] = manually_downloaded_files[
          'train_annotations'
      ]
    else:  # Use the provided test split.
      test_data_range = None
      manually_downloaded_files['test_all_frames'] = (
          dl_manager.manual_dir / 'test_all_frames.zip'
      )
      manually_downloaded_files['test_annotations'] = (
          dl_manager.manual_dir / 'test.json'
      )

    extracted_files = dl_manager.extract(manually_downloaded_files)
    val_dir = 'train_all_frames' if val_data_range else 'valid_all_frames'
    test_dir = 'train_all_frames' if test_data_range else 'test_all_frames'

    return {
        tfds.Split.TRAIN: self._generate_examples(
            annotations=extracted_files['train_annotations'],
            all_frames=extracted_files['train_all_frames']
            / 'train_all_frames'
            / 'JPEGImages',
            video_range_to_use=train_data_range,
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            annotations=extracted_files['valid_annotations'],
            all_frames=extracted_files['valid_all_frames']
            / val_dir
            / 'JPEGImages',
            video_range_to_use=val_data_range,
        ),
        tfds.Split.TEST: self._generate_examples(
            annotations=extracted_files['test_annotations'],
            all_frames=extracted_files['test_all_frames']
            / test_dir
            / 'JPEGImages',
            video_range_to_use=test_data_range,
        ),
    }

  def _maybe_resize_video(self, frames_list):
    """Resizes the video depending on the build_config."""
    if self.builder_config.height is None:
      return frames_list  # Don't waste compute loading and resizing.
    resized_images = []
    cv2 = tfds.core.lazy_imports.cv2
    for frame in frames_list:
      with epath.Path(frame).open('rb') as f:
        image = tfds.core.lazy_imports.PIL_Image.open(f).convert('RGB')
        image = np.asarray(image)
      image = cv2.resize(
          image, (self.builder_config.width, self.builder_config.height)
      )
      resized_images.append(image)
    return resized_images

  def _generate_examples(
      self,
      annotations: epath.Path,
      all_frames: epath.Path,
      video_range_to_use: Optional[Tuple[int, int]] = None,
  ):
    beam = tfds.core.lazy_imports.apache_beam
    annotations = json.loads(annotations.read_text())
    video_id_to_tracks, videos = _build_annotations_index(annotations)
    height = self._builder_config.height
    width = self._builder_config.width
    only_frames_with_labels = self._builder_config.only_frames_with_labels
    data_example = {}

    def _frame_index(frame_filename):
      """Convert a video frame filename into a numerical index."""
      basename = os.path.basename(os.fspath(frame_filename))
      return int(basename.split('.')[0])

    def _process_example(video_id):
      """Process a single video into a data example."""
      video = videos[video_id]
      if only_frames_with_labels:
        frames_list = [all_frames / file for file in video['file_names']]
      else:
        video_dir = os.path.dirname(video['file_names'][0])
        video_directory = all_frames / video_dir
        frames_list = list(video_directory.glob('*'))
      frames_list = sorted(frames_list, key=_frame_index)
      data_example['metadata'] = _create_metadata(
          video, height, width, len(frames_list)
      )
      data_example['tracks'] = []
      track_annotations = video_id_to_tracks[video_id]
      for track in track_annotations:
        data_example['tracks'].append(
            _create_per_track_annotation(
                video, frames_list, track, height, width
            )
        )
      data_example['video'] = self._maybe_resize_video(frames_list)
      return data_example['metadata']['video_name'], data_example

    video_keys = list(videos.keys())
    if video_range_to_use is not None:
      video_keys = video_keys[video_range_to_use[0] : video_range_to_use[1]]

    return beam.Create(video_keys) | beam.Map(_process_example)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class to lazy load DatasetBuilder classes from legacy locations.

This module is only intended to be used until all TFDS `DatasetsBuilder`s are
defined under the "datasets/" directory and the legacy locations have been
deprecated (ie: a new major release).
"""
from typing import Text

from absl import logging
from tensorflow_datasets.core import registered


class LazyBuilderImport:
  """Lazy load DatasetBuilder from given name from legacy locations."""

  def __init__(self, dataset_name: Text):
    object.__setattr__(self, "_dataset_name", dataset_name)
    object.__setattr__(self, "_dataset_cls", None)

  def _get_builder_cls(self):
    cls = object.__getattribute__(self, "_dataset_cls")
    if not cls:
      builder_name = object.__getattribute__(self, "_dataset_name")
      logging.warning(
          (
              "DEPRECATED! Do not use a DatasetBuilder class directly, but"
              " call `tfds.builder_cls('%s')`."
          ),
          builder_name,
      )
      cls = registered.imported_builder_cls(builder_name)
      object.__setattr__(self, "_dataset_cls", cls)
    return cls

  def __getattribute__(self, attr_name):
    cls = object.__getattribute__(self, "_get_builder_cls")()
    return getattr(cls, attr_name)

  def __call__(self, *args, **kwargs):
    cls = object.__getattribute__(self, "_get_builder_cls")()
    return cls(*args, **kwargs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The SuperGLUE benchmark."""

from __future__ import annotations

import json
import os

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_SUPER_GLUE_CITATION = """\
@article{wang2019superglue,
  title={SuperGLUE: A Stickier Benchmark for General-Purpose Language Understanding Systems},
  author={Wang, Alex and Pruksachatkun, Yada and Nangia, Nikita and Singh, Amanpreet and Michael, Julian and Hill, Felix and Levy, Omer and Bowman, Samuel R},
  journal={arXiv preprint arXiv:1905.00537},
  year={2019}
}

Note that each SuperGLUE dataset has its own citation. Please see the source to
get the correct citation for each contained dataset.
"""

_GLUE_DESCRIPTION = """\
SuperGLUE (https://super.gluebenchmark.com/) is a new benchmark styled after
GLUE with a new set of more difficult language understanding tasks, improved
resources, and a new public leaderboard.

"""

_BOOLQ_DESCRIPTION = """\
BoolQ (Boolean Questions, Clark et al., 2019a) is a QA task where each example consists of a short
passage and a yes/no question about the passage. The questions are provided anonymously and
unsolicited by users of the Google search engine, and afterwards paired with a paragraph from a
Wikipedia article containing the answer. Following the original work, we evaluate with accuracy."""

_CB_DESCRIPTION = """\
The CommitmentBank (De Marneffe et al., 2019) is a corpus of short texts in which at least
one sentence contains an embedded clause. Each of these embedded clauses is annotated with the
degree to which we expect that the person who wrote the text is committed to the truth of the clause.
The resulting task framed as three-class textual entailment on examples that are drawn from the Wall
Street Journal, fiction from the British National Corpus, and Switchboard. Each example consists
of a premise containing an embedded clause and the corresponding hypothesis is the extraction of
that clause. We use a subset of the data that had inter-annotator agreement above 0.85. The data is
imbalanced (relatively fewer neutral examples), so we evaluate using accuracy and F1, where for
multi-class F1 we compute the unweighted average of the F1 per class."""

_COPA_DESCRIPTION = """\
The Choice Of Plausible Alternatives (COPA, Roemmele et al., 2011) dataset is a causal
reasoning task in which a system is given a premise sentence and two possible alternatives. The
system must choose the alternative which has the more plausible causal relationship with the premise.
The method used for the construction of the alternatives ensures that the task requires causal reasoning
to solve. Examples either deal with alternative possible causes or alternative possible effects of the
premise sentence, accompanied by a simple question disambiguating between the two instance
types for the model. All examples are handcrafted and focus on topics from online blogs and a
photography-related encyclopedia. Following the recommendation of the authors, we evaluate using
accuracy."""

_RECORD_DESCRIPTION = """\
(Reading Comprehension with Commonsense Reasoning Dataset, Zhang et al., 2018) is a
multiple-choice QA task. Each example consists of a news article and a Cloze-style question about
the article in which one entity is masked out. The system must predict the masked out entity from a
given list of possible entities in the provided passage, where the same entity may be expressed using
multiple different surface forms, all of which are considered correct. Articles are drawn from CNN
and Daily Mail. Following the original work, we evaluate with max (over all mentions) token-level
F1 and exact match (EM)."""

_RTE_DESCRIPTION = """\
The Recognizing Textual Entailment (RTE) datasets come from a series of annual competitions
on textual entailment, the problem of predicting whether a given premise sentence entails a given
hypothesis sentence (also known as natural language inference, NLI). RTE was previously included
in GLUE, and we use the same data and format as before: We merge data from RTE1 (Dagan
et al., 2006), RTE2 (Bar Haim et al., 2006), RTE3 (Giampiccolo et al., 2007), and RTE5 (Bentivogli
et al., 2009). All datasets are combined and converted to two-class classification: entailment and
not_entailment. Of all the GLUE tasks, RTE was among those that benefited from transfer learning
the most, jumping from near random-chance performance (~56%) at the time of GLUE's launch to
85% accuracy (Liu et al., 2019c) at the time of writing. Given the eight point gap with respect to
human performance, however, the task is not yet solved by machines, and we expect the remaining
gap to be difficult to close."""

_MULTIRC_DESCRIPTION = """\
The Multi-Sentence Reading Comprehension dataset (MultiRC, Khashabi et al., 2018)
is a true/false question-answering task. Each example consists of a context paragraph, a question
about that paragraph, and a list of possible answers to that question which must be labeled as true or
false. Question-answering (QA) is a popular problem with many datasets. We use MultiRC because
of a number of desirable properties: (i) each question can have multiple possible correct answers,
so each question-answer pair must be evaluated independent of other pairs, (ii) the questions are
designed such that answering each question requires drawing facts from multiple context sentences,
and (iii) the question-answer pair format more closely matches the API of other SuperGLUE tasks
than span-based extractive QA does. The paragraphs are drawn from seven domains including news,
fiction, and historical text."""

_WIC_DESCRIPTION = """\
The Word-in-Context (WiC, Pilehvar and Camacho-Collados, 2019) dataset supports a word
sense disambiguation task cast as binary classification over sentence pairs. Given two sentences and a
polysemous (sense-ambiguous) word that appears in both sentences, the task is to determine whether
the word is used with the same sense in both sentences. Sentences are drawn from WordNet (Miller,
1995), VerbNet (Schuler, 2005), and Wiktionary. We follow the original work and evaluate using
accuracy."""

_WSC_DESCRIPTION = """\
The Winograd Schema Challenge (WSC, Levesque et al., 2012) is a reading comprehension
task in which a system must read a sentence with a pronoun and select the referent of that pronoun
from a list of choices. Given the difficulty of this task and the headroom still left, we have included
WSC in SuperGLUE and recast the dataset into its coreference form. The task is cast as a binary
classification problem, as opposed to N-multiple choice, in order to isolate the model's ability to
understand the coreference links within a sentence as opposed to various other strategies that may
come into play in multiple choice conditions. With that in mind, we create a split with 65% negative
majority class in the validation set, reflecting the distribution of the hidden test set, and 52% negative
class in the training set. The training and validation examples are drawn from the original Winograd
Schema dataset (Levesque et al., 2012), as well as those distributed by the affiliated organization
Commonsense Reasoning. The test examples are derived from fiction books and have been shared
with us by the authors of the original dataset. Previously, a version of WSC recast as NLI as included
in GLUE, known as WNLI. No substantial progress was made on WNLI, with many submissions
opting to submit only majority class predictions. WNLI was made especially difficult due to an
adversarial train/dev split: Premise sentences that appeared in the training set sometimes appeared
in the development set with a different hypothesis and a flipped label. If a system memorized the
training set without meaningfully generalizing, which was easy due to the small size of the training
set, it could perform far below chance on the development set. We remove this adversarial design
in the SuperGLUE version of WSC by ensuring that no sentences are shared between the training,
validation, and test sets.

However, the validation and test sets come from different domains, with the validation set consisting
of ambiguous examples such that changing one non-noun phrase word will change the coreference
dependencies in the sentence. The test set consists only of more straightforward examples, with a
high number of noun phrases (and thus more choices for the model), but low to no ambiguity."""

_AXB_DESCRIPTION = """\
An expert-constructed,
diagnostic dataset that automatically tests models for a broad range of linguistic, commonsense, and
world knowledge. Each example in this broad-coverage diagnostic is a sentence pair labeled with
a three-way entailment relation (entailment, neutral, or contradiction) and tagged with labels that
indicate the phenomena that characterize the relationship between the two sentences. Submissions
to the GLUE leaderboard are required to include predictions from the submission's MultiNLI
classifier on the diagnostic dataset, and analyses of the results were shown alongside the main
leaderboard. Since this broad-coverage diagnostic task has proved difficult for top models, we retain
it in SuperGLUE. However, since MultiNLI is not part of SuperGLUE, we collapse contradiction
and neutral into a single not_entailment label, and request that submissions include predictions
on the resulting set from the model used for the RTE task.
"""

_AXG_DESCRIPTION = """\
Winogender is designed to measure gender
bias in coreference resolution systems. We use the Diverse Natural Language Inference Collection
(DNC; Poliak et al., 2018) version that casts Winogender as a textual entailment task. Each example
consists of a premise sentence with a male or female pronoun and a hypothesis giving a possible
antecedent of the pronoun. Examples occur in minimal pairs, where the only difference between
an example and its pair is the gender of the pronoun in the premise. Performance on Winogender
is measured with both accuracy and the gender parity score: the percentage of minimal pairs for
which the predictions are the same. We note that a system can trivially obtain a perfect gender parity
score by guessing the same class for all examples, so a high gender parity score is meaningless unless
accompanied by high accuracy. As a diagnostic test of gender bias, we view the schemas as having high
positive predictive value and low negative predictive value; that is, they may demonstrate the presence
of gender bias in a system, but not prove its absence.
"""

_BOOLQ_CITATION = """\
@inproceedings{clark2019boolq,
  title={BoolQ: Exploring the Surprising Difficulty of Natural Yes/No Questions},
  author={Clark, Christopher and Lee, Kenton and Chang, Ming-Wei, and Kwiatkowski, Tom and Collins, Michael, and Toutanova, Kristina},
  booktitle={NAACL},
  year={2019}
}"""

_CB_CITATION = """\
@article{de marneff_simons_tonhauser_2019,
  title={The CommitmentBank: Investigating projection in naturally occurring discourse},
  journal={proceedings of Sinn und Bedeutung 23},
  author={De Marneff, Marie-Catherine and Simons, Mandy and Tonhauser, Judith},
  year={2019}
}"""

_COPA_CITATION = """\
@inproceedings{roemmele2011choice,
  title={Choice of plausible alternatives: An evaluation of commonsense causal reasoning},
  author={Roemmele, Melissa and Bejan, Cosmin Adrian and Gordon, Andrew S},
  booktitle={2011 AAAI Spring Symposium Series},
  year={2011}
}"""

_RECORD_CITATION = """\
@article{zhang2018record,
  title={Record: Bridging the gap between human and machine commonsense reading comprehension},
  author={Zhang, Sheng and Liu, Xiaodong and Liu, Jingjing and Gao, Jianfeng and Duh, Kevin and Van Durme, Benjamin},
  journal={arXiv preprint arXiv:1810.12885},
  year={2018}
}"""

_RTE_CITATION = """\
@inproceedings{dagan2005pascal,
  title={The PASCAL recognising textual entailment challenge},
  author={Dagan, Ido and Glickman, Oren and Magnini, Bernardo},
  booktitle={Machine Learning Challenges Workshop},
  pages={177--190},
  year={2005},
  organization={Springer}
}
@inproceedings{bar2006second,
  title={The second pascal recognising textual entailment challenge},
  author={Bar-Haim, Roy and Dagan, Ido and Dolan, Bill and Ferro, Lisa and Giampiccolo, Danilo and Magnini, Bernardo and Szpektor, Idan},
  booktitle={Proceedings of the second PASCAL challenges workshop on recognising textual entailment},
  volume={6},
  number={1},
  pages={6--4},
  year={2006},
  organization={Venice}
}
@inproceedings{giampiccolo2007third,
  title={The third pascal recognizing textual entailment challenge},
  author={Giampiccolo, Danilo and Magnini, Bernardo and Dagan, Ido and Dolan, Bill},
  booktitle={Proceedings of the ACL-PASCAL workshop on textual entailment and paraphrasing},
  pages={1--9},
  year={2007},
  organization={Association for Computational Linguistics}
}
@inproceedings{bentivogli2009fifth,
  title={The Fifth PASCAL Recognizing Textual Entailment Challenge.},
  author={Bentivogli, Luisa and Clark, Peter and Dagan, Ido and Giampiccolo, Danilo},
  booktitle={TAC},
  year={2009}
}"""

_MULTIRC_CITATION = """\
@inproceedings{MultiRC2018,
    author = {Daniel Khashabi and Snigdha Chaturvedi and Michael Roth and Shyam Upadhyay and Dan Roth},
    title = {Looking Beyond the Surface:A Challenge Set for Reading Comprehension over Multiple Sentences},
    booktitle = {Proceedings of North American Chapter of the Association for Computational Linguistics (NAACL)},
    year = {2018}
}"""

_WIC_CITATION = """\
@article{DBLP:journals/corr/abs-1808-09121,
  author={Mohammad Taher Pilehvar and os{\'{e}} Camacho{-}Collados},
  title={WiC: 10, 000 Example Pairs for Evaluating Context-Sensitive Representations},
  journal={CoRR},
  volume={abs/1808.09121},
  year={2018},
  url={http://arxiv.org/abs/1808.09121},
  archivePrefix={arXiv},
  eprint={1808.09121},
  timestamp={Mon, 03 Sep 2018 13:36:40 +0200},
  biburl={https://dblp.org/rec/bib/journals/corr/abs-1808-09121},
  bibsource={dblp computer science bibliography, https://dblp.org}
}"""

_WSC_CITATION = """\
@inproceedings{levesque2012winograd,
  title={The winograd schema challenge},
  author={Levesque, Hector and Davis, Ernest and Morgenstern, Leora},
  booktitle={Thirteenth International Conference on the Principles of Knowledge Representation and Reasoning},
  year={2012}
}"""

_AXG_CITATION = """\
@inproceedings{rudinger-EtAl:2018:N18,
  author    = {Rudinger, Rachel  and  Naradowsky, Jason  and  Leonard, Brian  and  {Van Durme}, Benjamin},
  title     = {Gender Bias in Coreference Resolution},
  booktitle = {Proceedings of the 2018 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies},
  month     = {June},
  year      = {2018},
  address   = {New Orleans, Louisiana},
  publisher = {Association for Computational Linguistics}
}
"""


class SuperGlueConfig(tfds.core.BuilderConfig):
  """BuilderConfig for SuperGLUE."""

  def __init__(
      self,
      *,
      features,
      data_url,
      citation,
      url,
      label_classes=("False", "True"),
      **kwargs,
  ):
    """BuilderConfig for SuperGLUE.

    Args:
      features: `list[string]`, list of the features that will appear in the
        feature dict. Should not include "label".
      data_url: `string`, url to download the zip file from.
      citation: `string`, citation for the data set.
      url: `string`, url for information about the data set.
      label_classes: `list[string]`, the list of classes for the label if the
        label is present as a string. Non-string labels will be cast to either
        'False' or 'True'.
      **kwargs: keyword arguments forwarded to super.
    """
    # Version history:
    # 1.0.2: Fixed non-nondeterminism in ReCoRD.
    # 1.0.1: Change from the pre-release trial version of SuperGLUE (v1.9) to
    #        the full release (v2.0).
    # 1.0.0: S3 (new shuffling, sharding and slicing mechanism).
    # 0.0.2: Initial version.
    super(SuperGlueConfig, self).__init__(
        version=tfds.core.Version("1.0.2"), **kwargs
    )
    self.features = features
    self.label_classes = label_classes
    self.data_url = data_url
    self.citation = citation
    self.url = url


class SuperGlue(tfds.core.GeneratorBasedBuilder):
  """The SuperGLUE benchmark."""

  BUILDER_CONFIGS = [
      SuperGlueConfig(
          name="boolq",
          description=_BOOLQ_DESCRIPTION,
          features=["question", "passage"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/BoolQ.zip"
          ),
          citation=_BOOLQ_CITATION,
          url="https://github.com/google-research-datasets/boolean-questions",
      ),
      SuperGlueConfig(
          name="cb",
          description=_CB_DESCRIPTION,
          features=["premise", "hypothesis"],
          label_classes=["entailment", "contradiction", "neutral"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/CB.zip"
          ),
          citation=_CB_CITATION,
          url="https://github.com/mcdm/CommitmentBank",
      ),
      SuperGlueConfig(
          name="copa",
          description=_COPA_DESCRIPTION,
          label_classes=["choice1", "choice2"],
          # Note that question will only be the X in the statement "What's
          # the X for this?".
          features=["premise", "choice1", "choice2", "question"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/COPA.zip"
          ),
          citation=_COPA_CITATION,
          url="http://people.ict.usc.edu/~gordon/copa.html",
      ),
      SuperGlueConfig(
          name="multirc",
          description=_MULTIRC_DESCRIPTION,
          features=["paragraph", "question", "answer"],
          data_url="https://dl.fbaipublicfiles.com/glue/superglue/data/v2/MultiRC.zip",
          citation=_MULTIRC_CITATION,
          url="https://cogcomp.org/multirc/",
      ),
      SuperGlueConfig(
          name="record",
          description=_RECORD_DESCRIPTION,
          # Note that entities and answers will be a sequences of strings. Query
          # will contain @placeholder as a substring, which represents the word
          # to be substituted in.
          features=["passage", "query", "entities", "answers"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/ReCoRD.zip"
          ),
          citation=_RECORD_CITATION,
          url="https://sheng-z.github.io/ReCoRD-explorer/",
      ),
      SuperGlueConfig(
          name="rte",
          description=_RTE_DESCRIPTION,
          features=["premise", "hypothesis"],
          label_classes=["entailment", "not_entailment"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/RTE.zip"
          ),
          citation=_RTE_CITATION,
          url="https://aclweb.org/aclwiki/Recognizing_Textual_Entailment",
      ),
      SuperGlueConfig(
          name="wic",
          description=_WIC_DESCRIPTION,
          # Note that start1, start2, end1, and end2 will be integers stored as
          # tf.int32.
          features=[
              "word",
              "sentence1",
              "sentence2",
              "start1",
              "start2",
              "end1",
              "end2",
          ],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/WiC.zip"
          ),
          citation=_WIC_CITATION,
          url="https://pilehvar.github.io/wic/",
      ),
      SuperGlueConfig(
          name="wsc",
          description=_WSC_DESCRIPTION,
          # Note that span1_index and span2_index will be integers stored as
          # tf.int32.
          features=[
              "text",
              "span1_index",
              "span2_index",
              "span1_text",
              "span2_text",
          ],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/WSC.zip"
          ),
          citation=_WSC_CITATION,
          url=(
              "https://cs.nyu.edu/faculty/davise/papers/WinogradSchemas/WS.html"
          ),
      ),
      SuperGlueConfig(
          name="wsc.fixed",
          description=(
              _WSC_DESCRIPTION
              + "\n\nThis version fixes issues where the spans are not"
              " actually "
              "substrings of the text."
          ),
          # Note that span1_index and span2_index will be integers stored as
          # tf.int32.
          features=[
              "text",
              "span1_index",
              "span2_index",
              "span1_text",
              "span2_text",
          ],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/WSC.zip"
          ),
          citation=_WSC_CITATION,
          url=(
              "https://cs.nyu.edu/faculty/davise/papers/WinogradSchemas/WS.html"
          ),
      ),
      SuperGlueConfig(
          name="axb",
          description=_AXB_DESCRIPTION,
          features=["sentence1", "sentence2"],
          label_classes=["entailment", "not_entailment"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/AX-b.zip"
          ),
          citation="",  # The GLUE citation is sufficient.
          url="https://gluebenchmark.com/diagnostics",
      ),
      SuperGlueConfig(
          name="axg",
          description=_AXG_DESCRIPTION,
          features=["premise", "hypothesis"],
          label_classes=["entailment", "not_entailment"],
          data_url=(
              "https://dl.fbaipublicfiles.com/glue/superglue/data/v2/AX-g.zip"
          ),
          citation=_AXG_CITATION,
          url="https://github.com/rudinger/winogender-schemas",
      ),
  ]

  def _info(self):
    features = {
        feature: tfds.features.Text()
        for feature in self.builder_config.features
    }
    if self.builder_config.name.startswith("wsc"):
      features["span1_index"] = tf.int32
      features["span2_index"] = tf.int32
    if self.builder_config.name == "wic":
      features["start1"] = tf.int32
      features["start2"] = tf.int32
      features["end1"] = tf.int32
      features["end2"] = tf.int32
    if self.builder_config.name == "multirc":
      features["idx"] = tfds.features.FeaturesDict({
          "paragraph": np.int32,
          "question": np.int32,
          "answer": np.int32,
      })
    elif self.builder_config.name == "record":
      features["idx"] = tfds.features.FeaturesDict({
          "passage": np.int32,
          "query": np.int32,
      })
    else:
      features["idx"] = tf.int32

    if self.builder_config.name == "record":
      # Entities are the set of possible choices for the placeholder.
      features["entities"] = tfds.features.Sequence(tfds.features.Text())
      # Answers are the subset of entities that are correct.
      features["answers"] = tfds.features.Sequence(tfds.features.Text())
    else:
      features["label"] = tfds.features.ClassLabel(
          names=self.builder_config.label_classes
      )

    return tfds.core.DatasetInfo(
        builder=self,
        description=_GLUE_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        homepage=self.builder_config.url,
        citation=self.builder_config.citation + "\n" + _SUPER_GLUE_CITATION,
    )

  def _split_generators(self, dl_manager):
    dl_dir = dl_manager.download_and_extract(self.builder_config.data_url) or ""
    task_name = _get_task_name_from_data_url(self.builder_config.data_url)
    dl_dir = os.path.join(dl_dir, task_name)
    if self.builder_config.name in ["axb", "axg"]:
      return [
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "data_file": os.path.join(
                      dl_dir, "{}.jsonl".format(task_name)
                  ),
                  "split": tfds.Split.TEST,
              },
          ),
      ]
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "data_file": os.path.join(dl_dir, "train.jsonl"),
                "split": tfds.Split.TRAIN,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "data_file": os.path.join(dl_dir, "val.jsonl"),
                "split": tfds.Split.VALIDATION,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "data_file": os.path.join(dl_dir, "test.jsonl"),
                "split": tfds.Split.TEST,
            },
        ),
    ]

  def _generate_examples(self, data_file, split):
    with epath.Path(data_file).open() as f:
      for line in f:
        row = json.loads(line)

        if self.builder_config.name == "multirc":
          paragraph = row["passage"]
          for question in paragraph["questions"]:
            for answer in question["answers"]:
              label = answer.get("label")
              key = "%s_%s_%s" % (row["idx"], question["idx"], answer["idx"])
              yield key, {
                  "paragraph": paragraph["text"],
                  "question": question["question"],
                  "answer": answer["text"],
                  "label": -1 if label is None else _cast_label(bool(label)),
                  "idx": {
                      "paragraph": row["idx"],
                      "question": question["idx"],
                      "answer": answer["idx"],
                  },
              }
        elif self.builder_config.name == "record":
          passage = row["passage"]
          for qa in row["qas"]:
            yield qa["idx"], {
                "passage": passage["text"],
                "query": qa["query"],
                "entities": _get_record_entities(passage),
                "answers": _get_record_answers(qa),
                "idx": {"passage": row["idx"], "query": qa["idx"]},
            }
        else:
          if self.builder_config.name.startswith("wsc"):
            row.update(row["target"])
          example = {
              feature: row[feature] for feature in self.builder_config.features
          }
          if self.builder_config.name == "wsc.fixed":
            example = _fix_wst(example)
          example["idx"] = row["idx"]

          if "label" in row:
            if self.builder_config.name == "copa":
              example["label"] = "choice2" if row["label"] else "choice1"
            else:
              example["label"] = _cast_label(row["label"])
          else:
            assert split == tfds.Split.TEST, row
            example["label"] = -1
          yield example["idx"], example


def _fix_wst(ex):
  """Fixes most cases where spans are not actually substrings of text."""

  def _fix_span_text(k):
    """Fixes a single span."""
    text = ex[k + "_text"]
    index = ex[k + "_index"]

    if text in ex["text"]:
      return

    if text in ("Kamenev and Zinoviev", "Kamenev, Zinoviev, and Stalin"):
      # There is no way to correct these examples since the subjects have
      # intervening text.
      return

    if "theyscold" in text:
      ex["text"].replace("theyscold", "they scold")  # pytype: disable=attribute-error
      ex["span2_index"] = 10
    # Make sure case of the first words match.
    first_word = ex["text"].split()[index]  # pytype: disable=attribute-error
    if first_word[0].islower():
      text = text[0].lower() + text[1:]
    else:
      text = text[0].upper() + text[1:]
    # Remove punctuation in span.
    text = text.rstrip(".")
    # Replace incorrect whitespace character in span.
    text = text.replace("\n", " ")
    ex[k + "_text"] = text
    assert ex[k + "_text"] in ex["text"], ex

  _fix_span_text("span1")
  _fix_span_text("span2")
  return ex


def _cast_label(label):
  """Converts the label into the appropriate string version."""
  if isinstance(label, str):
    return label
  elif isinstance(label, bool):
    return "True" if label else "False"
  elif isinstance(label, int):
    assert label in (0, 1)
    return str(label)
  else:
    raise ValueError(f"Invalid label format: {type(label)}")


def _get_record_entities(passage):
  """Returns the unique set of entities."""
  text = passage["text"]
  entities = set()
  for entity in passage["entities"]:
    entities.add(text[entity["start"] : entity["end"] + 1])
  return sorted(entities)


def _get_record_answers(qa):
  """Returns the unique set of answers."""
  if "answers" not in qa:
    return []
  answers = set()
  for answer in qa["answers"]:
    answers.add(answer["text"])
  return sorted(answers)


def _get_task_name_from_data_url(data_url):
  return data_url.split("/")[-1].split(".")[0]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machine Translations of XNLI: The Cross-Lingual NLI Corpus."""

import collections
import csv

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@article{hu2020xtreme,
      author    = {Junjie Hu and Sebastian Ruder and Aditya Siddhant and Graham Neubig and Orhan Firat and Melvin Johnson},
      title     = {XTREME: A Massively Multilingual Multi-task Benchmark for Evaluating Cross-lingual Generalization},
      journal   = {CoRR},
      volume    = {abs/2003.11080},
      year      = {2020},
      archivePrefix = {arXiv},
      eprint    = {2003.11080}
}"""

_DESCRIPTION = """\
This dataset contains machine translations of MNLI into each of the XNLI
languages. The translation data is provided by XTREME. Note that this is
different from the machine translated data provided by the original XNLI paper.
"""

_XTREME_TRANSLATIONS_FORMAT = 'https://storage.googleapis.com/xtreme_translations/XNLI/translate-train/en-{0}-translated.tsv'

_LANGUAGES = (
    'ar',
    'bg',
    'de',
    'el',
    'en',
    'es',
    'fr',
    'hi',
    'ru',
    'sw',
    'th',
    'tr',
    'ur',
    'vi',
    'zh',
)


class XtremeXnli(tfds.core.GeneratorBasedBuilder):
  """Machine translations of the MNLI dataset."""

  VERSION = tfds.core.Version('1.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'premise': tfds.features.Translation(
                languages=_LANGUAGES,
            ),
            'hypothesis': tfds.features.TranslationVariableLanguages(
                languages=_LANGUAGES,
            ),
            'label': tfds.features.ClassLabel(
                names=['entailment', 'neutral', 'contradiction']
            ),
        }),
        # No default supervised_keys (as we have to pass both premise
        # and hypothesis as input).
        supervised_keys=None,
        homepage='https://www.nyu.edu/projects/bowman/xnli/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    # Add urls pointing to machine translations of MNLI, provided by XTREME
    download_urls = {
        lang: _XTREME_TRANSLATIONS_FORMAT.format(lang)
        for lang in _LANGUAGES
        if lang != 'en'
    }
    dl_dirs = dl_manager.download_and_extract(download_urls)
    return {'train': self._generate_examples(dl_dirs)}

  def _generate_examples(self, dl_dirs):
    rows_per_pair_id = collections.defaultdict(list)
    en_pair_to_id = {}
    for lang in _LANGUAGES:
      # Translations are only for non-English languages.
      if lang == 'en':
        continue
      with tf.io.gfile.GFile(dl_dirs[lang]) as f:
        reader = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
        for row in reader:
          en_sent1, en_sent2, translated_sent1, translated_sent2, label = row
          # Change XTREME's label to conform with eval data.
          if label == 'contradictory':
            label = 'contradiction'
          key = (en_sent1, en_sent2)
          if key not in en_pair_to_id:
            en_pair_to_id[key] = len(en_pair_to_id)
            rows_per_pair_id[key].append({
                'language': 'en',
                'sentence1': en_sent1,
                'sentence2': en_sent2,
                'gold_label': label,
                'pairID': en_pair_to_id[key],
            })
          pair_id = en_pair_to_id[key]
          row = {
              'language': lang,
              'sentence1': translated_sent1,
              'sentence2': translated_sent2,
              'gold_label': label,
              'pairID': pair_id,
          }
          rows_per_pair_id[key].append(row)

    for rows in rows_per_pair_id.values():
      premise = {row['language']: row['sentence1'] for row in rows}
      hypothesis = {row['language']: row['sentence2'] for row in rows}
      yield rows[0]['pairID'], {
          'premise': premise,
          'hypothesis': hypothesis,
          'label': rows[0]['gold_label'],
      }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `tensorflow_datasets.core.visualization.show_examples`."""

from unittest import mock

import numpy as np
from tensorflow_datasets import testing
from tensorflow_datasets.core import load
from tensorflow_datasets.core import visualization

# Import for registration
from tensorflow_datasets.image_classification import imagenet  # pylint: disable=unused-import,g-bad-import-order


class ShowExamplesTest(testing.TestCase):

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples(self, mock_fig):
    with testing.mock_data(num_examples=20):
      ds, ds_info = load.load('imagenet2012', split='train', with_info=True)
    visualization.show_examples(ds, ds_info)

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples_supervised(self, _):
    with testing.mock_data(num_examples=20):
      ds, ds_info = load.load(
          'imagenet2012', split='train', with_info=True, as_supervised=True
      )
    visualization.show_examples(ds, ds_info)

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples_with_batch(self, _):
    with testing.mock_data(num_examples=20):
      ds, ds_info = load.load(
          'imagenet2012', split='train', with_info=True, batch_size=32
      )
    visualization.show_examples(ds, ds_info, is_batched=True)

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples_graph(self, _):
    with testing.mock_data(num_examples=20):
      ds, ds_info = load.load('ogbg_molpcba', split='train', with_info=True)
    visualization.show_examples(ds, ds_info)

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples_graph_with_colors_and_labels(self, _):
    with testing.mock_data(num_examples=20):
      ds, ds_info = load.load('ogbg_molpcba', split='train', with_info=True)

    # Dictionaries used to map nodes and edges to colors.
    atomic_numbers_to_elements = {
        6: 'C',
        7: 'N',
        8: 'O',
        9: 'F',
        14: 'Si',
        15: 'P',
        16: 'S',
        17: 'Cl',
        35: 'Br,',
    }
    elements_to_colors = {
        element: f'C{index}'
        for index, element in enumerate(atomic_numbers_to_elements.values())
    }
    bond_types_to_colors = {num: f'C{num}' for num in range(4)}

    # Node colors are atomic numbers.
    def node_color_fn(graph):
      atomic_numbers = 1 + graph['node_feat'][:, 0].numpy()
      return {
          index: elements_to_colors[atomic_numbers_to_elements[atomic_number]]
          for index, atomic_number in enumerate(atomic_numbers)
      }

    # Node labels are element names.
    def node_label_fn(graph):
      atomic_numbers = 1 + graph['node_feat'][:, 0].numpy()
      return {
          index: atomic_numbers_to_elements[atomic_number]
          for index, atomic_number in enumerate(atomic_numbers)
      }

    # Edge colors are bond types.
    def edge_color_fn(graph):
      bonds = graph['edge_index'].numpy()
      bond_types = graph['edge_feat'][:, 0].numpy()
      return {
          tuple(bond): bond_types_to_colors[bond_type]
          for bond, bond_type in zip(bonds, bond_types)
      }

    visualization.show_examples(
        ds,
        ds_info,
        node_color_fn=node_color_fn,
        node_label_fn=node_label_fn,
        edge_color_fn=edge_color_fn,
    )

  @mock.patch('matplotlib.pyplot.figure')
  def test_show_examples_missing_sample(self, _):
    with testing.mock_data(num_examples=3):
      ds, ds_info = load.load('imagenet2012', split='train', with_info=True)
    visualization.show_examples(ds.take(3), ds_info)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""GRef dataset.
"""
from __future__ import annotations

import collections
import json

import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The Google RefExp dataset is a collection of text descriptions of objects in
images which builds on the publicly available MS-COCO dataset. Whereas the
image captions in MS-COCO apply to the entire image, this dataset focuses on
text descriptions that allow one to uniquely identify a single object or region
within an image. See more details in this paper: Generation and Comprehension
of Unambiguous Object Descriptions.
"""

_CITATION = """
@inproceedings{mao2016generation,
  title={Generation and Comprehension of Unambiguous Object Descriptions},
  author={Mao, Junhua and Huang, Jonathan and Toshev, Alexander and Camburu, Oana and Yuille, Alan and Murphy, Kevin},
  booktitle={CVPR},
  year={2016}
}
"""


def _build_bbox(image_info, x, y, width, height):
  return tfds.features.BBox(
      ymin=y / image_info['height'],
      xmin=x / image_info['width'],
      ymax=(y + height) / image_info['height'],
      xmax=(x + width) / image_info['width'],
  )


def _combine_gref_anns(gref_anns):
  """Reorganize G-Ref by matching annotations to image_id.

  Also map referring expressions to annotation id.
  Args:
    gref_anns: a dict of gref_anns annotations.

  Returns:
    combine_anns: a list of dict based on gref_anns['images'].
      Each dict has a key 'instances' corresponding to a list of annotations
      of that image.
  """
  img_id2annotations = collections.defaultdict(list)
  for _, v in gref_anns['annotations'].items():
    # Add ref expression to annotation.
    # this adds the referent, the raw string, tokens, and the referring
    # expression id to the annotation.
    ref_list = []
    for rexp_id in sorted(v['refexp_ids']):
      r = gref_anns['refexps'][str(rexp_id)]
      if 'word' in r['parse']['referent']:
        r['referent'] = r['parse']['referent']['word']
      else:
        r['referent'] = ''
      r.pop('parse')
      ref_list.append(r)

    v['refexps'] = ref_list
    img_id2annotations[int(v['image_id'])].append(v)

  img_id2annotations = {
      k: list(sorted(v, key=lambda a: a['annotation_id']))
      for k, v in img_id2annotations.items()
  }
  combined_anns = []
  for _, v in gref_anns['images'].items():
    v['instances'] = img_id2annotations[v['image_id']]
    combined_anns.append((v['image_id'], v))

  combined_anns = sorted(combined_anns, key=lambda a: a[0])
  return combined_anns


class Gref(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for gref dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  MANUAL_DOWNLOAD_INSTRUCTIONS = """
    Follow instructions at https://github.com/mjhucla/Google_Refexp_toolbox
    to download and pre-process the data into aligned format with COCO.
    The directory contains 2 files and one folder:
    * google_refexp_train_201511_coco_aligned_catg.json
    * google_refexp_val_201511_coco_aligned_catg.json
    * coco_train2014/

    The coco_train2014 folder contains all of COCO 2014 training images.
    """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        # Description and homepage used for documentation
        description=_DESCRIPTION,
        homepage='https://github.com/mjhucla/Google_Refexp_toolbox',
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'image/id': np.int64,
            'objects': tfds.features.Sequence({
                'id': np.int64,
                'area': np.int64,
                'bbox': tfds.features.BBoxFeature(),
                'label': np.int64,
                'label_name': tfds.features.ClassLabel(num_classes=80),
                'refexp': tfds.features.Sequence({
                    'refexp_id': np.int64,
                    'tokens': tfds.features.Sequence(tfds.features.Text()),
                    'referent': tfds.features.Text(),
                    'raw': tfds.features.Text(),
                }),
            }),
        }),
        # If there's a common `(input, target)` tuple from the features,
        # specify them here. They'll be used if as_supervised=True in
        # builder.as_dataset.
        supervised_keys=None,
        # Bibtex citation for the dataset
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    coco_image_dir = dl_manager.manual_dir / 'coco_train2014'
    gref_aligned_json_train = (
        dl_manager.manual_dir
        / 'google_refexp_train_201511_coco_aligned_catg.json'
    )
    gref_aligned_json_val = (
        dl_manager.manual_dir
        / 'google_refexp_val_201511_coco_aligned_catg.json'
    )
    # Specify the splits
    return {
        tfds.Split.TRAIN: self._generate_examples(
            coco_image_dir,
            gref_aligned_json_train,
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            coco_image_dir,
            gref_aligned_json_val,
        ),
    }

  def _generate_examples(self, image_dir, gref_aligned_json):
    """Yields examples."""
    gref_anns = json.loads(gref_aligned_json.read_text())

    categories_id2name = {c['id']: c['name'] for c in gref_anns['categories']}
    self.info.features['objects']['label_name'].names = [
        c['name'] for c in gref_anns['categories']
    ]
    gref_anns = _combine_gref_anns(gref_anns)
    for image_id, image_ann in gref_anns:
      image_path = image_dir / image_ann['file_name']
      example = {
          'image': image_path,
          'image/id': image_id,
          'objects': [],
      }
      for inst in image_ann['instances']:
        # build instance dictionary
        example['objects'].append({
            'id': inst['annotation_id'],
            'area': inst['area'],
            'bbox': _build_bbox(image_ann, *inst['bbox']),
            'label': inst['category_id'],  # 1-index.
            'label_name': categories_id2name[inst['category_id']],
            'refexp': inst['refexps'],
        })
      yield image_id, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""robosuite_panda_pick_place_can dataset."""

from __future__ import annotations


import numpy as np
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rlds import rlds_base


_DESCRIPTION = """
These datasets have been created with the PickPlaceCan environment of the
[robosuite robotic arm simulator](https://robosuite.ai/). The human datasets
were recorded by a single operator using
the [RLDS Creator](https://github.com/google-research/rlds-creator) and a
gamepad controller.

The synthetic datasets have been recorded using the
[EnvLogger library](https://github.com/deepmind/envlogger).

The datasets follow the [RLDS format](https://github.com/google-research/rlds)
to represent steps and episodes.

Episodes consist of 400 steps. In each episode, a tag is
added when the task is completed, this tag is stored as part of the custom step
metadata.

Note that, due to the EnvLogger dependency, generation of this dataset is
currently supported on Linux environments only.
"""

_CITATION = """
@misc{ramos2021rlds,
      title={RLDS: an Ecosystem to Generate, Share and Use Datasets in Reinforcement Learning},
      author={Sabela Ramos and Sertan Girgin and Léonard Hussenot and Damien Vincent and Hanna Yakubovich and Daniel Toyama and Anita Gergely and Piotr Stanczyk and Raphael Marinier and Jeremiah Harmsen and Olivier Pietquin and Nikola Momchev},
      year={2021},
      eprint={2111.02767},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}
"""

_HOMEPAGE = 'https://github.com/google-research/rlds'


class RobosuitePandaPickPlaceCan(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for robosuite_panda_pick_place_can dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  _DATA_PATHS = {
      'human_dc29b40a': 'https://storage.googleapis.com/rlds_external_data_release/rlds_robosuite_panda_pick_place_can_human_state_only_dc29b40a.tar.gz',
      'human_images_dc29b40a': 'https://storage.googleapis.com/rlds_external_data_release/rlds_robosuite_panda_pick_place_can_human_dc29b40a.tar.gz',
      'synthetic_stochastic_sac_afe13968': 'https://storage.googleapis.com/rlds_external_data_release/rlds_robosuite_panda_pick_place_can_synthetic_stochastic_sac_afe13968.tar.gz',
  }

  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [
      rlds_base.DatasetConfig(
          name='human_dc29b40a',
          observation_info={
              'object-state': tfds.features.Tensor(
                  shape=(14,), dtype=np.float64
              ),
              'Can_pos': tfds.features.Tensor(shape=(3,), dtype=np.float64),
              'Can_to_robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float32
              ),
              'robot0_joint_pos_cos': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'robot0_gripper_qpos': tfds.features.Tensor(
                  shape=(2,), dtype=np.float64
              ),
              'robot0_proprio-state': tfds.features.Tensor(
                  shape=(32,), dtype=np.float64
              ),
              'robot0_joint_vel': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'robot0_joint_pos_sin': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'Can_quat': tfds.features.Tensor(shape=(4,), dtype=np.float64),
              'Can_to_robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float64
              ),
              'robot0_gripper_qvel': tfds.features.Tensor(
                  shape=(2,), dtype=np.float64
              ),
              'robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float64
              ),
              'robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float64
              ),
          },
          action_info=tfds.features.Tensor(shape=(7,), dtype=np.float64),
          reward_info=np.float64,
          discount_info=np.float64,
          episode_metadata_info={
              'agent_id': np.str_,
              'episode_index': np.int32,
              'episode_id': np.str_,
          },
          step_metadata_info={
              'tag:placed': np.bool_,
              'image': tfds.features.Image(),
          },
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description='Human generated dataset (50 episodes).',
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='human_images_dc29b40a',
          observation_info={
              'birdview_image': tfds.features.Image(
                  shape=(256, 256, 3), dtype=np.uint8, encoding_format='png'
              ),
              'object-state': tfds.features.Tensor(
                  shape=(14,), dtype=np.float64
              ),
              'Can_pos': tfds.features.Tensor(shape=(3,), dtype=np.float64),
              'Can_to_robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float32
              ),
              'robot0_joint_pos_cos': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'robot0_eye_in_hand_image': tfds.features.Image(
                  shape=(256, 256, 3), dtype=np.uint8, encoding_format='png'
              ),
              'robot0_gripper_qpos': tfds.features.Tensor(
                  shape=(2,), dtype=np.float64
              ),
              'robot0_proprio-state': tfds.features.Tensor(
                  shape=(32,), dtype=np.float64
              ),
              'robot0_robotview_image': tfds.features.Image(
                  shape=(256, 256, 3), dtype=np.uint8, encoding_format='png'
              ),
              'robot0_joint_vel': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'robot0_joint_pos_sin': tfds.features.Tensor(
                  shape=(7,), dtype=np.float64
              ),
              'agentview_image': tfds.features.Image(
                  shape=(256, 256, 3), dtype=np.uint8, encoding_format='png'
              ),
              'Can_quat': tfds.features.Tensor(shape=(4,), dtype=np.float64),
              'Can_to_robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float64
              ),
              'robot0_gripper_qvel': tfds.features.Tensor(
                  shape=(2,), dtype=np.float64
              ),
              'robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float64
              ),
              'robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float64
              ),
          },
          action_info=tfds.features.Tensor(shape=(7,), dtype=np.float64),
          reward_info=np.float64,
          discount_info=np.float64,
          episode_metadata_info={
              'agent_id': np.str_,
              'episode_index': np.int32,
              'episode_id': np.str_,
          },
          step_metadata_info={
              'tag:placed': np.bool_,
              'image': tfds.features.Image(),
          },
          citation=_CITATION,
          overall_description=_DESCRIPTION,
          description=(
              'Human generated dataset, including images with different camera'
              ' angles in the observation.'
              ' Note that it may take some time to generate.'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
      rlds_base.DatasetConfig(
          name='synthetic_stochastic_sac_afe13968',
          observation_info={
              'object-state': tfds.features.Tensor(
                  shape=(14,), dtype=np.float32
              ),
              'Can_pos': tfds.features.Tensor(shape=(3,), dtype=np.float32),
              'Can_to_robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float32
              ),
              'robot0_joint_pos_cos': tfds.features.Tensor(
                  shape=(7,), dtype=np.float32
              ),
              'robot0_gripper_qpos': tfds.features.Tensor(
                  shape=(2,), dtype=np.float32
              ),
              'robot0_proprio-state': tfds.features.Tensor(
                  shape=(32,), dtype=np.float32
              ),
              'robot0_joint_vel': tfds.features.Tensor(
                  shape=(7,), dtype=np.float32
              ),
              'robot0_joint_pos_sin': tfds.features.Tensor(
                  shape=(7,), dtype=np.float32
              ),
              'Can_quat': tfds.features.Tensor(shape=(4,), dtype=np.float32),
              'Can_to_robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float32
              ),
              'robot0_gripper_qvel': tfds.features.Tensor(
                  shape=(2,), dtype=np.float32
              ),
              'robot0_eef_pos': tfds.features.Tensor(
                  shape=(3,), dtype=np.float32
              ),
              'robot0_eef_quat': tfds.features.Tensor(
                  shape=(4,), dtype=np.float32
              ),
          },
          action_info=tfds.features.Tensor(shape=(7,), dtype=np.float32),
          reward_info=np.float64,
          discount_info=np.float64,
          episode_metadata_info={
              'agent_id': np.str_,
              'episode_index': np.int32,
              'episode_id': np.str_,
          },
          step_metadata_info={
              'tag:placed': np.bool_,
              'image': tfds.features.Image(),
          },
          citation=_CITATION,
          homepage=_HOMEPAGE,
          overall_description=_DESCRIPTION,
          description=(
              'Synthetic dataset generated by a stochastic agent trained with'
              ' SAC (200 episodes).'
          ),
          supervised_keys=None,  # pytype: disable=wrong-arg-types  # gen-stub-imports
      ),
  ]

  # pytype: enable=wrong-keyword-args


  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return rlds_base.build_info(self.builder_config, self)

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        {
            'file_path': self._DATA_PATHS[self.builder_config.name],
        }
    )
    return {
        'train': self._generate_examples(path),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    file_path = path['file_path']
    return rlds_base.generate_examples(file_path)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared utilities for QA datasets."""
from __future__ import annotations

import json

from absl import logging
from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds


def squadlike_features():
  return tfds.features.FeaturesDict({
      "id": np.str_,
      "title": tfds.features.Text(),
      "context": tfds.features.Text(),
      "question": tfds.features.Text(),
      "answers": tfds.features.Sequence({
          "text": tfds.features.Text(),
          "answer_start": np.int32,
      }),
  })


def generate_squadlike_examples(filepath):
  """Parses a SQuAD-like JSON, yielding examples with `squadlike_features`."""
  logging.info("generating examples from = %s", filepath)

  # We first re-group the answers, which may be flattened (e.g., by XTREME).
  qas = {}
  with epath.Path(filepath).open() as f:
    squad = json.load(f)
    for article in squad["data"]:
      title = article.get("title", "")
      for paragraph in article["paragraphs"]:
        context = paragraph["context"]
        for qa in paragraph["qas"]:
          qa["title"] = title
          qa["context"] = context
          id_ = qa["id"]
          if id_ in qas:
            qas[id_]["answers"].extend(qa["answers"])
          else:
            qas[id_] = qa

    for id_, qa in qas.items():
      answer_starts = [answer["answer_start"] for answer in qa["answers"]]
      answers = [answer["text"] for answer in qa["answers"]]
      yield id_, {
          "title": qa["title"],
          "context": qa["context"],
          "question": qa["question"],
          "id": id_,
          "answers": {
              "answer_start": answer_starts,
              "text": answers,
          },
      }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""grain tests."""

from unittest import mock

import pytest
import tensorflow_datasets as tfds
from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.data_sources import array_record
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_data_source


_FILE_INSTRUCTIONS = [
    shard_utils.FileInstruction(
        'my_file-000-of-003', skip=0, take=12, examples_in_shard=12
    ),
    shard_utils.FileInstruction(
        'my_file-001-of-003', skip=2, take=9, examples_in_shard=11
    ),
    shard_utils.FileInstruction(
        'my_file-002-of-003', skip=0, take=4, examples_in_shard=4
    ),
]


def create_dataset_info():
  with mock.patch.object(splits_lib, 'SplitInfo') as split_mock:
    split_mock.return_value.name = 'train'
    split_mock.return_value.file_instructions = _FILE_INSTRUCTIONS
    dataset_info = mock.create_autospec(dataset_info_lib.DatasetInfo)
    dataset_info.file_format = file_adapters.FileFormat.ARRAY_RECORD
    dataset_info.splits = {'train': split_mock()}
    dataset_info.name = 'dataset_name'
    return dataset_info


@pytest.mark.parametrize(
    ['file_format'],
    [
        (file_adapters.FileFormat.RIEGELI,),
        (file_adapters.FileFormat.TFRECORD,),
    ],
)
def test_unsupported_file_formats_raise_error(file_format):
  dataset_info = create_dataset_info()
  dataset_info.file_format = file_format
  with pytest.raises(
      NotImplementedError,
      match='No random access data source for file format',
  ):
    array_record.ArrayRecordDataSource(dataset_info, split='train')


def test_missing_split_raises_error():
  dataset_info = create_dataset_info()
  with pytest.raises(
      ValueError,
      match="Unknown split 'doesnotexist'.",
  ):
    array_record.ArrayRecordDataSource(dataset_info, split='doesnotexist')


def test_array_record_file_format_delegates_to_array_record_data_source():
  dataset_info = create_dataset_info()
  with mock.patch.object(
      array_record_data_source, 'ArrayRecordDataSource'
  ) as array_record_data_source_mock:
    array_record.ArrayRecordDataSource(dataset_info, split='train')
    array_record_data_source_mock.assert_called_once_with(_FILE_INSTRUCTIONS)


def test_data_source_calls_array_record_data_source():
  dataset_info = create_dataset_info()
  mock_data_source = mock.create_autospec(
      array_record_data_source.ArrayRecordDataSource,
      spec_set=True,
  )
  mock_data_source.__getitem__.side_effect = [
      'serialized example',
  ]
  mock_data_source.__len__.return_value = 42
  with mock.patch.object(
      array_record_data_source,
      'ArrayRecordDataSource',
      return_value=mock_data_source,
  ), mock.patch.object(
      dataset_info.features,
      'deserialize_example_np',
      return_value='deserialized example',
  ) as deserialize_example_mock:
    data_source = array_record.ArrayRecordDataSource(
        dataset_info, split='train'
    )
    assert data_source[12] == 'deserialized example'
    deserialize_example_mock.assert_called_with(
        'serialized example', decoders=None
    )
    assert deserialize_example_mock.call_count == 1


def test_repr_returns_meaningful_string_without_decoders():
  dataset_info = create_dataset_info()
  with mock.patch.object(array_record_data_source, 'ArrayRecordDataSource'):
    source = array_record.ArrayRecordDataSource(dataset_info, split='train')
    assert (
        repr(source)
        == "DataSource(name=dataset_name, split='train', decoders=None)"
    )


def test_repr_returns_meaningful_string_with_decoders():
  dataset_info = create_dataset_info()
  with mock.patch.object(array_record_data_source, 'ArrayRecordDataSource'):
    source = array_record.ArrayRecordDataSource(
        dataset_info,
        split='train',
        decoders={'my_feature': decode.SkipDecoding()},
    )
    assert (
        repr(source)
        == 'DataSource(name=dataset_name,'
        " split='train', decoders={'my_feature': <class"
        " 'tensorflow_datasets.core.decode.base.SkipDecoding'>})"
    )


def test_data_source_is_iterable():
  dataset_info = create_dataset_info()
  mock_data_source = mock.create_autospec(
      array_record_data_source.ArrayRecordDataSource,
      spec_set=True,
  )
  mock_data_source.__getitem__.return_value = [1, 2, 3]
  mock_data_source.__len__.return_value = 3
  with mock.patch.object(
      array_record_data_source,
      'ArrayRecordDataSource',
      return_value=mock_data_source,
  ), mock.patch.object(
      dataset_info.features,
      'deserialize_example_np',
      return_value='deserialized example',
  ) as deserialize_example_mock:
    data_source = array_record.ArrayRecordDataSource(
        dataset_info, split='train'
    )
    assert len(data_source) == 3
    for _ in data_source:
      continue
    assert mock_data_source.__getitem__.call_count == 3
    assert deserialize_example_mock.call_count == 3
    assert mock_data_source.__getitem__.call_args_list[0].args[0] == 0
    assert mock_data_source.__getitem__.call_args_list[1].args[0] == 1
    assert mock_data_source.__getitem__.call_args_list[2].args[0] == 2


def test_data_source_is_sliceable():
  mock_array_record_data_source = tfds.testing.PickableDataSourceMock()
  with tfds.testing.mock_data(
      mock_array_record_data_source=mock_array_record_data_source
  ):
    tfds.data_source('mnist', split='train')
    assert len(mock_array_record_data_source.call_args_list) == 1
    file_instructions = mock_array_record_data_source.call_args_list[0].args[0]
    assert file_instructions[0].skip == 0
    assert file_instructions[0].take == 60000

    tfds.data_source('mnist', split='train[:50%]')
    assert len(mock_array_record_data_source.call_args_list) == 2
    file_instructions = mock_array_record_data_source.call_args_list[1].args[0]
    assert file_instructions[0].skip == 0
    assert file_instructions[0].take == 30000



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dataset_collection_builder_testing."""
from tensorflow_datasets import testing
from tensorflow_datasets.testing import dataset_collection_builder_testing as dc_testing


def test_dc_test_base():
  class TestDummyCollection(dc_testing.DatasetCollectionTestBase):
    DATASET_COLLECTION_CLASS = testing.DummyDatasetCollection

  # Check defaults class attributes are set up correctly.
  assert TestDummyCollection.VERSION is None
  assert not TestDummyCollection.DATASETS_TO_TEST
  assert TestDummyCollection.CHECK_DATASETS_VERSION



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Removes metadata that are not present in the registered versions of TFDS.

Instructions:

```
python -m tensorflow_datasets.scripts.cleanup.metadata_cleanup
```


"""

import os
from typing import List

from absl import app
from absl import flags
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

FLAGS = flags.FLAGS

flags.DEFINE_boolean('dry_run', True, 'Dry run')
flags.DEFINE_string(
    'tfds_dir',
    str(tfds.core.utils.tfds_write_path()),
    'Path to tensorflow_datasets directory',
)


def _extract_metadata_versions(metadata_dir: str) -> List[str]:
  """Get all metadata directory versions paths.

  It only extracts the paths like 'dataset_name/version'
  or 'dataset_name/config/versions' in metadata dir.

  Args:
    metadata_dir: Path to metadata directory (testing/metadata).

  Returns:
    Existing metadata full names.
  """
  existing_names = []
  for root, _, _ in tf.io.gfile.walk(metadata_dir):
    full_name = root[len(metadata_dir) + 1 :]
    if tfds.core.load.is_full_name(full_name):
      existing_names.append(full_name)
  return existing_names


def _delete_metadata_dirs(metadata_dir: str) -> None:
  """Removes metadata that are not present in the registered versions of TFDS.

  Args:
    metadata_dir: Path to metadata directory (testing/metadata).
  """
  registered_names = set(tfds.core.load.list_full_names())
  existing_names = set(_extract_metadata_versions(metadata_dir))
  for extra_full_name in sorted(existing_names - registered_names):
    path_to_delete = os.path.join(metadata_dir, extra_full_name)
    print(f'Delete: {path_to_delete}')
    if FLAGS.dry_run:
      continue
    tf.io.gfile.rmtree(path_to_delete)


def main(_):
  """Main script."""
  # Delete metadata versions not present in register version.
  metadata_dir = os.path.join(FLAGS.tfds_dir, 'testing', 'metadata')
  _delete_metadata_dirs(metadata_dir)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorFlow utils."""

from __future__ import annotations

import collections
import contextlib
from typing import Any, Union

import numpy as np
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

# Struct containing a graph for the TFGraphRunner
GraphRun = collections.namedtuple(
    'GraphRun', 'graph, session, placeholder, output'
)

# Struct containing the run args, kwargs
RunArgs = collections.namedtuple('RunArgs', 'fct, input')


class TFGraphRunner(object):
  """Run in session mode or Eager mode.

  This is a compatibility util between graph and eager TensorFlow.

  The graph runner allow to run function defining small TensorFlow graphs:
   * In eager mode: The function is simply run eagerly and the result is
     returned
   * In graph mode: The first time, the function is compiled in a new graph,
     then, every time the same function will be called, the cached graph and
     session will be run.

  Ideally, one graph runner should only be used with a single function to avoid
  having too many opened session in session mode.
  Limitations:
   * Currently the graph runner only support function with single input
     and output. Support for more complex function could be added and should be
     relatively straightforward.
   * A different graph is created for each input shape, so it isn't really
     adapted for dynamic batch size.

  Usage:
    graph_runner = TFGraphRunner()
    output = graph_runner.run(tf.sigmoid, np.ones(shape=(5,)))
  """

  __slots__ = ['_graph_run_cache']

  def __init__(self):
    """Constructor."""
    # Cache containing all compiled graph and opened session. Only used in
    # non-eager mode.
    self._graph_run_cache = {}

  def __getstate__(self):
    return {}

  def __setstate__(self, state):
    self.__init__(**state)

  def run(self, fct, input_):
    """Execute the given TensorFlow function."""
    # TF 2.0
    if tf.executing_eagerly():
      return fct(input_).numpy()
    # TF 1.0
    else:
      # Should compile the function if this is the first time encountered
      if not isinstance(input_, np.ndarray):
        input_ = np.array(input_)
      run_args = RunArgs(fct=fct, input=input_)
      signature = self._build_signature(run_args)
      if signature not in self._graph_run_cache:
        graph_run = self._build_graph_run(run_args)
        self._graph_run_cache[signature] = graph_run
      else:
        graph_run = self._graph_run_cache[signature]

      # Then execute the cached graph
      return graph_run.session.run(
          graph_run.output,
          feed_dict={graph_run.placeholder: input_},
      )

  def _build_graph_run(self, run_args):
    """Create a new graph for the given args."""
    # Could try to use tfe.py_func(fct) but this would require knowing
    # information about the signature of the function.

    # Create a new graph:
    with tf.Graph().as_default() as g:
      # Create placeholder
      input_ = run_args.input
      placeholder = tf.compat.v1.placeholder(
          dtype=input_.dtype, shape=input_.shape
      )
      output = run_args.fct(placeholder)
    return GraphRun(
        session=raw_nogpu_session(g),
        graph=g,
        placeholder=placeholder,
        output=output,
    )

  def _build_signature(self, run_args):
    """Create a unique signature for each fct/inputs."""
    return (id(run_args.fct), run_args.input.dtype, run_args.input.shape)

  def __del__(self):
    # Close all sessions
    for graph_run in self._graph_run_cache.values():
      graph_run.session.close()


def convert_to_shape(shape: Any) -> type_utils.Shape:
  """Converts a shape to a TFDS shape."""
  if isinstance(shape, tuple):
    return shape
  if isinstance(shape, tf.TensorShape):
    return tuple(shape.as_list())
  if isinstance(shape, list):
    return tuple(shape)
  raise ValueError(
      f'Shape of type {type(shape)} with content {shape} is not supported!'
  )


@py_utils.memoize(maxsize=1000)
def assert_shape_match(
    shape1: type_utils.Shape, shape2: type_utils.Shape
) -> None:
  """Ensure shape1 matches the pattern given by shape2.

  Ex:
    assert_shape_match((64, 64, 3), (None, None, 3))

  Args:
    shape1 (tuple): Static shape
    shape2 (tuple): Dynamic shape (can contain None)
  """
  if shape1 is None or shape2 is None:
    if shape1 != shape2:
      raise ValueError(f'Shapes {shape1} and {shape2} must have the same rank')
    return
  rank1 = len(shape1)
  rank2 = len(shape2)
  if rank1 != rank2:
    raise ValueError(f'Shapes {shape1} and {shape2} must have the same rank')
  for dimension1, dimension2 in zip(shape1, shape2):
    if dimension1 is None or dimension2 is None:
      continue
    if dimension1 != dimension2:
      raise ValueError(f'Shapes {shape1} and {shape2} are incompatible')


def assert_tf_shape_match(
    shape1: tf.TensorShape, shape2: tf.TensorShape
) -> None:
  if shape1.rank is None or shape2.rank is None:
    raise ValueError(
        'Shapes must have known rank. Got %s and %s.'
        % (shape1.rank, shape2.rank)
    )
  shape1.assert_same_rank(shape2)
  shape1.assert_is_compatible_with(shape2)


def shapes_are_compatible(
    shapes0: type_utils.TreeDict[type_utils.Shape],
    shapes1: type_utils.TreeDict[type_utils.Shape],
) -> bool:
  """Returns True if all shapes are compatible."""
  # Use `py_utils.map_nested` instead of `tree_utils.map_structure` as shapes
  # are tuple/list.
  shapes0 = py_utils.map_nested(tf.TensorShape, shapes0, dict_only=True)
  shapes1 = py_utils.map_nested(tf.TensorShape, shapes1, dict_only=True)
  all_values = tree_utils.map_structure(
      lambda s0, s1: s0.is_compatible_with(s1),
      shapes0,
      shapes1,
  )
  return all(tf.nest.flatten(all_values))


def normalize_shape(
    shape: Union[type_utils.Shape, tf.TensorShape]
) -> type_utils.Shape:
  """Normalize `tf.TensorShape` to tuple of int/None."""
  if isinstance(shape, tf.TensorShape):
    return tuple(shape.as_list())  # pytype: disable=attribute-error
  else:
    assert isinstance(shape, tuple)
    return shape


def merge_shape(
    tf_shape: Union[tf.Tensor, np.ndarray], np_shape: type_utils.Shape
):
  """Returns the most static version of the shape.

  Static `None` values are replaced by dynamic `tf.Tensor` values.

  Example:

  ```
  merge_shape(
      tf_shape=tf.constant([28, 28, 3]),
      np_shape=(None, None, 3),
  ) == (tf.Tensor(numpy=28), tf.Tensor(numpy=28), 3)
  ```

  Args:
    tf_shape: The tf.Tensor containing the shape (e.g. `tf.shape(x)`)
    np_shape: The static shape tuple (e.g. `(None, None, 3)`)

  Returns:
    A tuple like np_shape, but with `None` values replaced by `tf.Tensor` values
  """
  assert_shape_match(tf_shape.shape, (len(np_shape),))
  return tuple(
      tf_shape[i] if dim is None else dim for i, dim in enumerate(np_shape)
  )


@contextlib.contextmanager
def nogpu_session(graph=None):
  """tf.Session context manager, hiding GPUs."""
  # We don't use the with construction because we don't want the Session to be
  # installed as the "default" session.
  sess = raw_nogpu_session(graph)
  yield sess
  sess.close()


def raw_nogpu_session(graph=None):
  """tf.Session, hiding GPUs."""
  config = tf.compat.v1.ConfigProto(device_count={'GPU': 0})
  return tf.compat.v1.Session(config=config, graph=graph)


@contextlib.contextmanager
def maybe_with_graph(graph=None, create_if_none=True):
  """Eager-compatible Graph().as_default() yielding the graph."""
  if tf.executing_eagerly():
    yield None
  else:
    if graph is None and create_if_none:
      graph = tf.Graph()

    if graph is None:
      yield None
    else:
      with graph.as_default():
        yield graph



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.dataset_builder."""

import pathlib
from typing import Callable
from unittest import mock

import apache_beam as beam
from etils import epath
import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import download
from tensorflow_datasets.core import features
from tensorflow_datasets.core import utils


class DummyBeamDataset(dataset_builder.GeneratorBasedBuilder):
  VERSION = utils.Version('1.0.0')

  EXPECTED_METADATA = {
      'valid_1000': 1000,
      'valid_725': 725,
  }

  def _info(self):
    return dataset_info.DatasetInfo(
        builder=self,
        features=features.FeaturesDict({
            'image': features.Image(shape=(16, 16, 1)),
            'label': features.ClassLabel(names=['dog', 'cat']),
            'id': tf.int32,
        }),
        supervised_keys=('x', 'x'),
        metadata=dataset_info.BeamMetadataDict(),
    )

  def _split_generators(self, dl_manager):
    del dl_manager
    return {
        'train': self._generate_examples(num_examples=1000),
        'test': self._generate_examples(num_examples=725),
    }

  def _generate_examples(self, num_examples):
    """Generate examples as dicts."""
    examples = beam.Create(range(num_examples)) | beam.Map(_gen_example)

    # Can save int, str,... metadata but not `beam.PTransform`
    self.info.metadata[f'valid_{num_examples}'] = num_examples
    with pytest.raises(
        NotImplementedError, match="can't be used on `beam.PTransform`"
    ):
      self.info.metadata[f'invalid_{num_examples}'] = _compute_sum(examples)
    return examples


class CommonPipelineDummyBeamDataset(DummyBeamDataset):
  EXPECTED_METADATA = {
      'label_sum_1000': 500,
      'id_mean_1000': 499.5,
      'label_sum_725': 362,
      'id_mean_725': 362.0,
  }

  def _split_generators(self, dl_manager, pipeline):
    del dl_manager

    examples = pipeline | beam.Create(range(1000)) | beam.Map(_gen_example)

    # Wrap the pipeline inside a ptransform_fn to add `'label' >> ` to avoid
    # duplicated PTransform nodes names.
    generate_examples = beam.ptransform_fn(self._generate_examples)
    return {
        'train': examples | 'train' >> generate_examples(num_examples=1000),
        'test': examples | 'test' >> generate_examples(num_examples=725),
    }

  def _generate_examples(self, examples, num_examples):
    """Generate examples as dicts."""
    examples |= beam.Filter(lambda x: x[0] < num_examples)
    # Record metadata works for common PCollections
    self.info.metadata[f'id_mean_{num_examples}'] = _compute_mean(examples)
    self.info.metadata[f'label_sum_{num_examples}'] = _compute_sum(examples)
    return examples


def _gen_example(x):
  return (
      x,
      {
          'image': (np.ones((16, 16, 1)) * x % 255).astype(np.uint8),
          'label': x % 2,
          'id': x,
      },
  )


def _compute_sum(examples):
  return (
      examples | beam.Map(lambda x: x[1]['label']) | beam.CombineGlobally(sum)
  )


def _compute_mean(examples):
  return (
      examples
      | beam.Map(lambda x: x[1]['id'])
      | beam.CombineGlobally(beam.combiners.MeanCombineFn())
  )


def make_default_config():
  return download.DownloadConfig()


@pytest.mark.parametrize(
    'dataset_cls', [DummyBeamDataset, CommonPipelineDummyBeamDataset]
)
@pytest.mark.parametrize(
    'make_dl_config',
    [
        make_default_config,
    ],
)
def test_beam_datasets(
    tmp_path: pathlib.Path,
    dataset_cls: dataset_builder.GeneratorBasedBuilder,
    make_dl_config: Callable[[], download.DownloadConfig],
):
  dataset_name = dataset_cls.name

  builder = dataset_cls(data_dir=tmp_path)
  builder.download_and_prepare(download_config=make_dl_config())

  data_path = tmp_path / dataset_name / '1.0.0'
  assert data_path.exists()  # Dataset has been generated

  # Check number of shards/generated files
  _test_shards(
      data_path,
      pattern='%s-test.tfrecord-{:05}-of-{:05}' % dataset_name,
      # Liquid sharding is not guaranteed to always use the same number.
      num_shards=builder.info.splits['test'].num_shards,
  )
  _test_shards(
      data_path,
      pattern='%s-train.tfrecord-{:05}-of-{:05}' % dataset_name,
      num_shards=1,
  )

  ds = dataset_utils.as_numpy(builder.as_dataset())

  def get_id(ex):
    return ex['id']

  _assert_values_equal(
      sorted(list(ds['test']), key=get_id),
      sorted([_gen_example(i)[1] for i in range(725)], key=get_id),
  )
  _assert_values_equal(
      sorted(list(ds['train']), key=get_id),
      sorted([_gen_example(i)[1] for i in range(1000)], key=get_id),
  )

  assert builder.info.metadata == builder.EXPECTED_METADATA


def _test_shards(data_path, pattern, num_shards):
  assert num_shards >= 1
  shards_filenames = [pattern.format(i, num_shards) for i in range(num_shards)]
  assert all(data_path.joinpath(f).exists() for f in shards_filenames)


def _assert_values_equal(nested_lhs, nested_rhs):
  """assertAllEqual applied to a list of nested elements."""
  for dict_lhs, dict_rhs in zip(nested_lhs, nested_rhs):
    flat_lhs = tf.nest.flatten(dict_lhs)
    flat_rhs = tf.nest.flatten(dict_rhs)
    for lhs, rhs in zip(flat_lhs, flat_rhs):
      np.testing.assert_array_equal(lhs, rhs)


def test_read_tfrecord_beam():
  builder = DummyBeamDataset()
  with mock.patch.object(
      beam.io, 'ReadFromTFRecord'
  ) as mock_read, mock.patch.object(epath, 'Path') as mock_epath:
    file_pattern = '/a/b/*'
    mock_epath.return_value.expanduser.return_value = file_pattern
    mock_epath.return_value.glob.return_value = ['/a/b/c', '/a/b/d']
    builder.read_tfrecord_beam(file_pattern, validate=True)
    mock_epath.return_value.glob.assert_called_once_with('a/b/*')
    mock_read.assert_called_once_with(file_pattern=file_pattern, validate=True)
    info_proto = builder.info.as_proto
    assert len(info_proto.data_source_accesses) == 2
    assert info_proto.data_source_accesses[0].file_system.path == '/a/b/c'
    assert info_proto.data_source_accesses[1].file_system.path == '/a/b/d'



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for write_metadata_lib."""

import pathlib

from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core.folder_dataset import write_metadata_utils

_PATH = epath.Path('/a/b')


@pytest.mark.parametrize(
    'file_format',
    [
        None,
    ],
)
def test_write_metadata(
    tmp_path: pathlib.Path,
    file_format,
):
  tmp_path = epath.Path(tmp_path)

  src_builder = testing.DummyDataset(
      data_dir=tmp_path / 'origin',
      file_format=file_format,
  )
  src_builder.download_and_prepare()

  dst_dir = tmp_path / 'copy'
  dst_dir.mkdir()

  # Copy all the tfrecord files, but not the dataset info
  for f in src_builder.data_path.iterdir():
    if naming.FilenameInfo.is_valid(f.name):
      f.copy(dst_dir / f.name)

  metadata_path = dst_dir / 'dataset_info.json'

  if file_format is None:
    split_infos = list(src_builder.info.splits.values())
  else:
    split_infos = None  # Auto-compute split infos

  assert not metadata_path.exists()
  write_metadata_utils.write_metadata(
      data_dir=dst_dir,
      features=src_builder.info.features,
      split_infos=split_infos,
      description='my test description.',
  )
  assert metadata_path.exists()

  # After metadata are written, builder can be restored from the directory
  builder = read_only_builder.builder_from_directory(dst_dir)
  assert builder.name == 'dummy_dataset'
  assert builder.version == '1.0.0'
  assert set(builder.info.splits) == {'train'}
  assert builder.info.splits['train'].num_examples == 3
  assert builder.info.description == 'my test description.'

  # Values are the same
  src_ds = src_builder.as_dataset(split='train')
  ds = builder.as_dataset(split='train')
  assert list(src_ds.as_numpy_iterator()) == list(ds.as_numpy_iterator())


@pytest.mark.parametrize(
    ['data_dir', 'filename_template', 'result'],
    [
        (
            _PATH,
            '{DATASET}',
            naming.ShardedFileTemplate(data_dir=_PATH, template='{DATASET}'),
        ),
        (
            _PATH,
            None,
            naming.ShardedFileTemplate(
                data_dir=_PATH,
                template='{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}',
            ),
        ),
        (
            _PATH,
            naming.ShardedFileTemplate(data_dir=epath.Path('/other')),
            naming.ShardedFileTemplate(
                data_dir=_PATH,
                template='{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}',
            ),
        ),
        (
            _PATH,
            naming.ShardedFileTemplate(
                data_dir=epath.Path('/other'), dataset_name='mnist'
            ),
            naming.ShardedFileTemplate(
                data_dir=_PATH,
                dataset_name='mnist',
                template='{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}',
            ),
        ),
    ],
)
def test_construct_filename_template(data_dir, filename_template, result):
  actual = write_metadata_utils._construct_filename_template(
      data_dir=data_dir, filename_template=filename_template
  )
  assert actual == result


@pytest.mark.parametrize(
    ['filename_template', 'file_infos', 'result'],
    [
        (
            naming.ShardedFileTemplate(data_dir=_PATH, template='{DATASET}'),
            [
                naming.FilenameInfo(
                    dataset_name='mnist', filetype_suffix='tfrecord'
                )
            ],
            naming.ShardedFileTemplate(
                data_dir=_PATH,
                template='{DATASET}',
                dataset_name='mnist',
                filetype_suffix='tfrecord',
            ),
        ),
        (
            naming.ShardedFileTemplate(
                data_dir=_PATH, template='{DATASET}', dataset_name='mnist'
            ),
            [naming.FilenameInfo(dataset_name='mnist')],
            naming.ShardedFileTemplate(
                data_dir=_PATH, template='{DATASET}', dataset_name='mnist'
            ),
        ),
        (
            naming.ShardedFileTemplate(data_dir=_PATH, template='{DATASET}'),
            [naming.FilenameInfo(filetype_suffix='tfrecord')],
            naming.ShardedFileTemplate(
                data_dir=_PATH, template='{DATASET}', filetype_suffix='tfrecord'
            ),
        ),
    ],
)
def test_enrich_filename_template(filename_template, file_infos, result):
  actual = write_metadata_utils._enrich_filename_template(
      filename_template=filename_template, file_infos=file_infos
  )
  assert actual == result


def test_enrich_filename_template_different_dataset_name():
  filename_template = naming.ShardedFileTemplate(
      data_dir=_PATH, template='{DATASET}', dataset_name='imagenet'
  )
  file_infos = [naming.FilenameInfo(dataset_name='mnist')]
  with pytest.raises(ValueError, match='Detected dataset name.+'):
    write_metadata_utils._enrich_filename_template(
        filename_template=filename_template, file_infos=file_infos
    )


def test_enrich_filename_template_different_filetype_suffix():
  filename_template = naming.ShardedFileTemplate(
      data_dir=_PATH, template='{DATASET}', filetype_suffix='tfrecord'
  )
  file_infos = [naming.FilenameInfo(filetype_suffix='riegeli')]
  with pytest.raises(ValueError, match='Detected filetype suffix.+'):
    write_metadata_utils._enrich_filename_template(
        filename_template=filename_template, file_infos=file_infos
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util to generate the dataset documentation content.

Used by tensorflow_datasets/scripts/documentation/build_catalog.py
"""

from concurrent import futures
import dataclasses
import functools
from typing import Any, Iterator, List, Optional, Set, Type

from absl import logging
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.scripts.documentation import collection_markdown_builder
from tensorflow_datasets.scripts.documentation import dataset_markdown_builder
from tensorflow_datasets.scripts.documentation import doc_utils
import tqdm

_WORKER_COUNT_DATASETS = 50
_WORKER_COUNT_CONFIGS = 20

# In the section headers we have different spellings of the same word. This
# mapping is used to convert to a single spelling.
_SECTION_WORD_MAPPING = {'modelling': 'modeling'}

# WmtTranslate: The raw wmt can only be instantiated with the config kwargs
_BUILDER_BLACKLIST = [
    'wmt_translate',
]

# pylint: disable=logging-format-interpolation


def _format_section(section: str) -> str:
  section = section.replace('_', ' ').replace('-', ' ').lower()
  for orig_word, new_word in _SECTION_WORD_MAPPING.items():
    section = section.replace(orig_word, new_word)
  return section.capitalize()


@dataclasses.dataclass(eq=False)
class BuilderToDocument:
  """Structure containing metadata."""

  sections: Set[str]
  namespace: Optional[str]
  builder: tfds.core.DatasetBuilder
  config_builders: List[tfds.core.DatasetBuilder]

  def __post_init__(self):
    # Make sure section names are formatted correctly.
    self.sections = {_format_section(section) for section in self.sections}


@dataclasses.dataclass(eq=False)
class BuilderDocumentation:
  """Documentation output of a single builder.

  Attributes:
    name: Documentation page name (e.g. `mnist`, `kaggle:mnist`)
    filestem: Documentation page name without suffix (e.g. `mnist`,
      `kaggle_mnist`)
    content: Documentation content
    sections: Documentation sections (e.g `text`, `image`,...)
    namespace: Dataset namespace
    is_manual: Whether the dataset require manual download
    is_nightly: Whether the dataset was recently added in `tfds-nightly`
  """

  name: str
  filestem: str
  content: str
  sections: Set[str]
  is_manual: bool
  is_nightly: bool

  def __post_init__(self):
    # Make sure section names are formatted correctly.
    self.sections = {_format_section(section) for section in self.sections}


@dataclasses.dataclass(eq=False, frozen=True)
class CollectionDocumentation:
  """Documentation output of a single builder.

  Attributes:
    name: Documentation page name (e.g. `xtreme`)
    content: Documentation content
    is_nightly: Whether the dataset was recently added in `tfds-nightly`
  """

  name: str
  content: str
  is_nightly: bool


def _load_builder(
    name: str,
) -> Optional[BuilderToDocument]:
  """Load the builder to document.

  Args:
    name: Builder to load

  Returns:
    BuilderToDocument or None.
  """
  try:
    dataset_name = tfds.core.naming.DatasetName(name)
  except ValueError as e:
    logging.warning('Could not parse dataset with name %s', name, exc_info=e)
    return None
  if dataset_name.namespace:  # Community dataset
    return _load_builder_from_location(name=name, dataset_name=dataset_name)
  else:  # Code dataset
    return _load_builder_from_code(name)


def _load_builder_from_location(
    name: str,
    dataset_name: tfds.core.naming.DatasetName,
) -> Optional[BuilderToDocument]:
  """Load the builder, config,... to document."""
  logging.info('Loading builder %s from location', dataset_name)
  try:
    builder = tfds.builder(name)
    logging.debug('Loaded builder from location: %s', builder)
  except tfds.core.DatasetNotFoundError as e:
    logging.info(
        'Dataset %s not found, will now try to load from sub-folder',
        dataset_name,
        exc_info=e,
    )
    # If tfds.core.DatasetNotFoundError, it might be the default
    # config isn't found. Should try to load a sub-folder (to check).
    builder = _maybe_load_config(name)
    if not builder:
      logging.error('Dataset %s not found', dataset_name, exc_info=e)
      return None
  except (OSError, tf.errors.PermissionDeniedError) as e:
    logging.error('Permission denied for %s', dataset_name, exc_info=e)
    tqdm.tqdm.write(f'Warning: Skip dataset {name} due to permission error')
    return None
  except Exception as e:  # pylint: disable=broad-except
    logging.error('CorruptedDatasetError: %s', repr(name), exc_info=e)
    return None
  if builder.builder_config:
    config_builders = _load_all_configs(name, builder)
  else:
    config_builders = []
  return BuilderToDocument(
      sections={dataset_name.namespace},
      namespace=dataset_name.namespace,
      builder=builder,
      config_builders=config_builders,
  )


def _maybe_load_config(name: str) -> Optional[tfds.core.DatasetBuilder]:
  del name  # unused
  return None


def _load_all_configs(
    name: str,
    builder: tfds.core.DatasetBuilder,
) -> List[tfds.core.DatasetBuilder]:
  """Load all builder configs."""
  # `data_dir/name/config/version/` -> `data_dir/name/`
  common_dir = builder.data_path.parent.parent
  filtered_dirs = {builder.builder_config.name, '.config'}

  def get_config_builder(path):
    if path.name in filtered_dirs:
      return None  # Default config is already loaded
    try:
      builder_conf = tfds.builder(f'{name}/{path.name}')
    except tfds.core.DatasetNotFoundError:
      return None
    if not builder_conf.builder_config:
      # Unexpected sub-config with wrong metadata.
      # This can happen if the user manually messed up with the directories.
      return None
    return builder_conf

  with futures.ThreadPoolExecutor(max_workers=_WORKER_COUNT_CONFIGS) as tpool:
    config_names = sorted(common_dir.iterdir())
    config_builders = [
        config_builder
        for config_builder in tpool.map(get_config_builder, config_names)
        if config_builder is not None  # Filter invalid configs
    ]

  return [builder] + config_builders


def _load_builder_from_code(
    name: str,
) -> BuilderToDocument:
  """Load the builder, config,... to document."""
  builder_cls = tfds.builder_cls(name)
  sections = _get_sections(builder_cls)

  if builder_cls.BUILDER_CONFIGS:  # Builder with configs

    def get_config_builder(config) -> tfds.core.DatasetBuilder:
      return tfds.builder(builder_cls.name, config=config)

    with futures.ThreadPoolExecutor(max_workers=_WORKER_COUNT_CONFIGS) as tpool:
      config_builders = list(
          tpool.map(get_config_builder, builder_cls.BUILDER_CONFIGS),
      )
    return BuilderToDocument(
        sections=sections,
        namespace=None,
        builder=config_builders[0],
        config_builders=config_builders,
    )
  else:  # Builder without configs
    return BuilderToDocument(
        sections=sections,
        namespace=None,
        builder=builder_cls(),  # pytype: disable=not-instantiable
        config_builders=[],
    )


def _get_sections(builder_cls: Type[tfds.core.DatasetBuilder]) -> Set[str]:
  """Returns the sections associated with the builder."""
  module_parts = builder_cls.__module__.split('.')
  if module_parts[0] != 'tensorflow_datasets':
    raise AssertionError(f'Unexpected builder {builder_cls}: module')
  # Legacy datasets: a single section, inferred from module path.
  _, category, *_ = module_parts  # tfds.<category>.xyz
  if category != 'datasets':
    return {category}
  # Sections are inferred from tags.
  ds_metadata = builder_cls.get_metadata()
  if ds_metadata.tags:
    sections = set()
    for tag in ds_metadata.tags:
      # Ignore languages until we have a better way to display them.
      if tag.startswith('content.language'):
        continue
      section = tag.rsplit('.')[-1]
      sections.add(section)
    return sections
  return {'uncategorized'}


def _document_single_collection(
    name: str,
) -> CollectionDocumentation:
  """Returns the documentation for a single dataset collection."""
  logging.info('Documenting dataset collection %s', name)
  collection = tfds.dataset_collection(name=name)
  out_str = collection_markdown_builder.get_collection_markdown_string(
      collection=collection
  )
  return CollectionDocumentation(
      name=name,
      content=out_str,
      # TODO(tfds): Add support for display of nightly icon when necessary.
      is_nightly=False,
  )


def _document_single_builder(
    name: str,
    **kwargs: Any,
) -> Optional[BuilderDocumentation]:
  """Doc string for a single builder, with or without configs."""
  with tfds.core.utils.try_reraise(f'Error for `{name}`: '):
    return _document_single_builder_inner(name, **kwargs)


def _document_single_builder_inner(
    name: str,
    nightly_doc_util: doc_utils.NightlyDocUtil,
    **kwargs: Any,
) -> Optional[BuilderDocumentation]:
  """Doc string for a single builder, with or without configs."""
  tqdm.tqdm.write(f'Document builder {name}...')
  doc_info = _load_builder(name)
  if doc_info is None:
    logging.warning(
        'No doc info was found for document builder %s. Skipping.', name
    )
    return None

  out_str = dataset_markdown_builder.get_markdown_string(
      namespace=doc_info.namespace,
      builder=doc_info.builder,
      config_builders=doc_info.config_builders,
      nightly_doc_util=nightly_doc_util,
      **kwargs,
  )
  is_nightly = bool(
      nightly_doc_util and nightly_doc_util.is_builder_nightly(name)
  )
  return BuilderDocumentation(
      name=name,
      filestem=name.replace(':', '_'),
      content=out_str,
      sections=doc_info.sections,
      is_manual=bool(doc_info.builder.MANUAL_DOWNLOAD_INSTRUCTIONS),
      is_nightly=is_nightly,
  )


def _all_tfds_datasets() -> List[str]:
  """Returns all "official" TFDS dataset names."""
  datasets = []
  for name in tfds.list_builders(with_community_datasets=True):
    if name in _BUILDER_BLACKLIST:
      continue
    try:
      _ = tfds.core.naming.parse_builder_name_kwargs(name=name)
      datasets.append(name)
    except ValueError as e:
      logging.error('Could not parse dataset with name `%s`!', name, exc_info=e)
  return sorted(datasets)


def iter_collections_documentation(
    collection_names: Optional[List[str]] = None,
) -> Iterator[CollectionDocumentation]:
  logging.info('Retrieving dataset collection names...')
  collection_names = collection_names or sorted(tfds.list_dataset_collections())
  for collection_name in collection_names:
    yield _document_single_collection(name=collection_name)
  print('All collections documentations generated!')


def iter_documentation_builders(
    datasets: Optional[List[str]] = None,
    *,
    doc_util_paths: Optional[doc_utils.DocUtilPaths] = None,
) -> Iterator[BuilderDocumentation]:
  """Creates dataset documentation string for given datasets.

  Args:
    datasets: list of datasets for which to create documentation. If None, then
      all available datasets will be used.
    doc_util_paths: Additional path for visualization, nightly info,...

  Yields:
    builder_documetation: The documentation information for each builder
  """
  print('Retrieving the list of builders...')
  datasets = datasets or _all_tfds_datasets()

  # pytype: disable=attribute-error
  if doc_util_paths.fig_base_path:
    visu_doc_util = doc_utils.VisualizationDocUtil(
        base_path=doc_util_paths.fig_base_path,
        base_url=doc_util_paths.fig_base_url,
    )
  else:
    visu_doc_util = None

  if doc_util_paths.df_base_path:
    df_doc_util = doc_utils.DataframeDocUtil(
        base_path=doc_util_paths.df_base_path,
        base_url=doc_util_paths.df_base_url,
    )
  else:
    df_doc_util = None

  if doc_util_paths.fig_base_path:
    nightly_doc_util = doc_utils.NightlyDocUtil(
        path=doc_util_paths.nightly_path,
    )
  else:
    nightly_doc_util = None
  # pytype: enable=attribute-error

  document_single_builder_fn = functools.partial(
      _document_single_builder,
      visu_doc_util=visu_doc_util,
      df_doc_util=df_doc_util,
      nightly_doc_util=nightly_doc_util,
  )

  # Document all builders
  print(f'Document {len(datasets)} builders...')
  with futures.ThreadPoolExecutor(max_workers=_WORKER_COUNT_DATASETS) as tpool:
    tasks = [
        tpool.submit(document_single_builder_fn, name) for name in datasets
    ]
    for future in tqdm.tqdm(futures.as_completed(tasks), total=len(tasks)):
      builder_doc = future.result()
      if builder_doc is None:  # Builder filtered
        continue
      else:
        tqdm.tqdm.write(f'Documentation generated for {builder_doc.name}...')
        yield builder_doc
  print('All builder documentations generated!')



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DatasetBuilder that stores data as a TFDS dataset.

Adhoc builders can be used to easily create a new TFDS dataset without having to
define a new `DatasetBuilder` class. This can be handy when working in a
notebook. For example, if you are in a notebook and you want to transform a
`tf.data.Dataset` into a TFDS dataset, then you can do so as follows:

```
import numpy as np

import tensorflow as tf
import tensorflow_datasets.public_api as tfds

my_ds_train = tf.data.Dataset.from_tensor_slices({"number": [1, 2, 3]})
my_ds_test = tf.data.Dataset.from_tensor_slices({"number": [4, 5]})

# Optionally define a custom `data_dir`. If None, then the default data dir is
# used.
custom_data_dir = "/my/folder"

builder = tfds.dataset_builders.store_as_tfds_dataset(
    name="my_dataset",
    config="single_number",
    version="1.0.0",
    data_dir=custom_data_dir,
    split_datasets={
        "train": my_ds_train,
        "test": my_ds_test,
    },
    features=tfds.features.FeaturesDict({
        "number": tfds.features.Scalar(dtype=np.int64),
    }),
    description="My dataset with a single number.",
    release_notes={
        "1.0.0": "Initial release with numbers up to 5!",
    }
)
```

The `config` argument is optional and can be useful if you want to store
different configs under the same TFDS dataset.

The `data_dir` argument can be used to store the generated TFDS dataset in a
different folder, for example in your own sandbox if you don't want to share
this with others (yet). Note that when doing this, you also need to pass the
`data_dir` to `tfds.load`. If the `data_dir` argument is not specified, then
the default TFDS data dir will be used.

After the TFDS dataset has been stored, it can be loaded from other scripts:

```
# If no custom data dir was specified:
ds_test = tfds.load("my_dataset/single_number", split="test")

# When there are multiple versions, you can also specify the version.
ds_test = tfds.load("my_dataset/single_number:1.0.0", split="test")

# If the TFDS was stored in a custom folder, then it can be loaded as follows:
custom_data_dir = "/my/folder"
ds_test = tfds.load("my_dataset/single_number:1.0.0", split="test",
data_dir=custom_data_dir)
```
"""
from __future__ import annotations

import sys
import typing
from typing import Any, Dict, Iterable, Mapping, Optional, Union

from absl import logging
from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import download
from tensorflow_datasets.core import features as feature_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

KeyExample = split_builder_lib.KeyExample

if typing.TYPE_CHECKING:
  import tensorflow as tf  # pytype: disable=import-error
  import apache_beam as beam  # pytype: disable=import-error

  BeamInput = Union[beam.PTransform, beam.PCollection[KeyExample]]
  InputData = Union[
      tf.data.Dataset,
      beam.PTransform,
      beam.PCollection[KeyExample],
      Iterable[KeyExample],
  ]
else:
  BeamInput = Union[
      "beam.PTransform", "beam.PCollection[split_builder_lib.KeyExample]"
  ]
  InputData = Union[
      "tf.data.Dataset",
      "beam.PTransform",
      "beam.PCollection[split_builder_lib.KeyExample]",
      "Iterable[KeyExample]",
  ]


class AdhocBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """Dataset builder that allows building a dataset without defining a class."""

  def __init__(
      self,
      *,
      name: str,
      version: Union[utils.Version, str],
      features: feature_lib.FeatureConnector,
      split_datasets: Mapping[str, InputData],
      config: Union[None, str, dataset_builder.BuilderConfig] = None,
      data_dir: Optional[epath.PathLike] = None,
      description: Optional[str] = None,
      release_notes: Optional[Mapping[str, str]] = None,
      homepage: Optional[str] = None,
      citation: Optional[str] = None,
      file_format: Optional[Union[str, file_adapters.FileFormat]] = None,
      disable_shuffling: Optional[bool] = False,
      **kwargs: Any,
  ):
    self.name = name
    self.VERSION = utils.Version(version)  # pylint: disable=invalid-name
    self.RELEASE_NOTES = release_notes  # pylint: disable=invalid-name
    if config:
      if isinstance(config, str):
        config = dataset_builder.BuilderConfig(
            name=config, version=version, release_notes=release_notes
        )
      self.BUILDER_CONFIGS = [config]  # pylint: disable=invalid-name
    self._split_datasets = split_datasets
    self._feature_spec = features
    self._description = (
        description or "Dataset built without a DatasetBuilder class."
    )
    self._homepage = homepage
    self._citation = citation
    self._disable_shuffling = disable_shuffling
    super().__init__(
        data_dir=data_dir,
        config=config,
        version=version,
        file_format=file_format,
        **kwargs,
    )

  def _info(self) -> dataset_info.DatasetInfo:
    return dataset_info.DatasetInfo(
        builder=self,
        description=self._description,
        features=self._feature_spec,
        homepage=self._homepage,
        citation=self._citation,
        disable_shuffling=self._disable_shuffling,
    )

  def _split_generators(
      self, dl_manager: download.DownloadManager
  ) -> Dict[splits_lib.Split, split_builder_lib.SplitGenerator]:
    del dl_manager
    split_generators = {}
    beam = sys.modules.get("apache_beam", None)
    for split_name, dataset in self._split_datasets.items():
      if isinstance(dataset, tf.data.Dataset):
        split_generators[split_name] = self._generate_examples_tf_data(dataset)
      elif isinstance(dataset, Iterable):
        split_generators[split_name] = self._generate_examples_iterator(dataset)
      elif beam and (
          isinstance(dataset, beam.PTransform)
          or isinstance(dataset, beam.PCollection)
      ):
        split_generators[split_name] = dataset
      else:
        raise ValueError(f"Dataset type {type(dataset)} not supported.")
    return split_generators

  def _generate_examples_tf_data(
      self, ds: tf.data.Dataset
  ) -> split_builder_lib.SplitGenerator:
    for i, example in enumerate(dataset_utils.as_numpy(ds)):
      yield i, example

  def _generate_examples_iterator(
      self,
      ds: Iterable[KeyExample],
  ) -> split_builder_lib.SplitGenerator:
    yield from ds

  def _generate_examples(
      self, **kwargs: Any
  ) -> split_builder_lib.SplitGenerator:
    raise NotImplementedError()


def store_as_tfds_dataset(
    name: str,
    version: Union[utils.Version, str],
    features: feature_lib.FeatureConnector,
    split_datasets: Mapping[str, InputData],
    config: Union[None, str, dataset_builder.BuilderConfig] = None,
    data_dir: Optional[epath.PathLike] = None,
    description: Optional[str] = None,
    release_notes: Optional[Mapping[str, str]] = None,
    homepage: Optional[str] = None,
    citation: Optional[str] = None,
    file_format: Optional[Union[str, file_adapters.FileFormat]] = None,
    download_config: Optional[download.DownloadConfig] = None,
    disable_shuffling: Optional[bool] = False,
) -> AdhocBuilder:
  """Store a dataset as a TFDS dataset."""
  if not split_datasets:
    raise ValueError("No splits with datasets were given.")
  ds_types = {type(ds) for ds in split_datasets.values()}
  if len(ds_types) > 1:
    raise TypeError(
        f"All split datasets should have the same type. Got: {ds_types}"
    )

  builder = AdhocBuilder(
      name=name,
      version=version,
      features=features,
      split_datasets=split_datasets,
      config=config,
      data_dir=data_dir,
      description=description,
      release_notes=release_notes,
      homepage=homepage,
      citation=citation,
      file_format=file_format,
      disable_shuffling=disable_shuffling,
  )
  builder.download_and_prepare(
      download_config=download_config, file_format=file_format
  )
  logging.info(
      "Dataset '%s' was prepared as a TFDS dataset in folder %s",
      name,
      builder.data_dir,
  )
  return builder


class TfDataBuilder(AdhocBuilder, skip_registration=True):
  """Builds datasets from tf.data.Dataset without defining a class.

  This is kept for backwards-compatibility.
  """

  def __init__(
      self,
      *,
      split_datasets: Mapping[str, tf.data.Dataset],
      **kwargs: Any,
  ):
    logging.warning(
        "This class is deprecated. "
        "Please use tfds.dataset_builders.store_as_tfds_dataset instead."
    )
    super().__init__(split_datasets=split_datasets, **kwargs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.scripts.cli.main."""

from unittest import mock

from tensorflow_datasets.scripts.cli import main


def test_main():
  def _check_exit(status=0, message=None):
    del message
    assert status == 0  # Check argparse exit gracefully

  # Argparse call `sys.exit(0)` when `--version` is passed.
  with mock.patch('sys.exit', _check_exit):
    version_flag = '--version'
    main.main(main._parse_flags(['', version_flag]))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mtnt dataset."""
import csv
import dataclasses
import hashlib

from absl import logging
from etils import epath
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
MTNT: Machine Translation of Noisy Text
"""

_CITATION = """
@InProceedings{michel2018mtnt,
  author    = {Michel, Paul  and  Neubig, Graham},
  title     = {MTNT: A Testbed for Machine Translation of Noisy Text},
  booktitle = {Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing}
}
"""

URL_1_1 = 'https://github.com/pmichel31415/mtnt/releases/download/v1.1/MTNT.1.1.tar.gz'


@dataclasses.dataclass
class MtntConfig(tfds.core.BuilderConfig):
  src_lang: str = ''
  dst_lang: str = ''


class Mtnt(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for mtnt dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      # pylint: disable=unexpected-keyword-arg
      MtntConfig(name='en-fr', src_lang='en', dst_lang='fr'),
      MtntConfig(name='en-ja', src_lang='en', dst_lang='ja'),
      MtntConfig(name='fr-en', src_lang='fr', dst_lang='en'),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    # TODO(mtnt): Specifies the tfds.core.DatasetInfo object
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            # These are the features of your dataset like images, labels ...
            'src': tfds.features.Text(),
            'dst': tfds.features.Text(),
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=('src', 'dst'),  # Set to `None` to disable
        homepage='https://pmichel31415.github.io/mtnt/index.html',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(URL_1_1)

    pair = f'{self.builder_config.src_lang}-{self.builder_config.dst_lang}'
    return {
        'train': self._generate_examples(path / f'MTNT/train/train.{pair}.tsv'),
        'test': self._generate_examples(path / f'MTNT/test/test.{pair}.tsv'),
        'valid': self._generate_examples(path / f'MTNT/valid/valid.{pair}.tsv'),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    with epath.Path(path).open() as f:
      reader = csv.reader(f, delimiter='\t')
      for row in reader:
        if len(row) < 3:
          logging.info('skipped row %s', row)
          continue
        key = hashlib.md5(','.join(row[0:3]).encode('utf-8')).hexdigest()
        yield key, {
            'src': row[1],
            'dst': row[2],
        }
